import matplotlib.pyplot as plt
import numpy as np

from orqviz.scans import perform_1D_scan, perform_2D_interpolation
from orqviz.scans.plots import plot_1D_scan_result, plot_2D_interpolation_result


def SUM_OF_SINES(params):
    return np.sum(np.sin(params))


def test_plot_1D_scan_result():
    origin = np.array([0.0, 0.0])
    direction_x = np.array([1.0, 0.0])
    end_points_x = (-np.pi, np.pi)
    n_steps_x = 100

    scan_1d = perform_1D_scan(
        origin=origin,
        loss_function=SUM_OF_SINES,
        direction=direction_x,
        n_steps=n_steps_x,
        end_points=end_points_x,
    )

    plot_1D_scan_result(scan_1d)
    fig, ax = plt.subplots()
    plot_1D_scan_result(scan_1d, ax)


def test_plot_2D_interpolation_result():
    point_1 = np.array([1, 0])
    point_2 = np.array([0, 2])
    end_points_x = (-0.5, 1.5)
    end_points_y = (-0.5, 0.5)
    n_steps_x = 100
    n_steps_y = 10

    scan_2d = perform_2D_interpolation(
        point_1=point_1,
        point_2=point_2,
        loss_function=SUM_OF_SINES,
        n_steps_x=n_steps_x,
        n_steps_y=n_steps_y,
        end_points_x=end_points_x,
        end_points_y=end_points_y,
    )

    plot_2D_interpolation_result(scan_2d)
    fig, ax = plt.subplots()
    plot_2D_interpolation_result(scan_2d, fig, ax)



from typing import Callable, Optional, Tuple

import numpy as np

from ..aliases import (
    ArrayOfParameterVectors,
    DirectionVector,
    LossFunction,
    ParameterVector,
)
from ..geometric import (
    direction_linspace,
    get_random_normal_vector,
    relative_periodic_wrap,
)
from .data_structures import Scan1DResult
from .evals import eval_points_on_path


def perform_1D_scan(
    origin: ParameterVector,
    loss_function: LossFunction,
    direction: Optional[DirectionVector] = None,
    n_steps: int = 31,
    end_points: Tuple[float, float] = (-np.pi, np.pi),
    verbose: bool = False,
) -> Scan1DResult:
    """Function to perform a 1D scan on a loss function in a specific direction.

    Args:
        origin: Parameter vector that is the origin on the 1D scan.
        loss_function: Function to perform the scan on. It must receive only a
            numpy.ndarray of parameters, and return a real number.
            If your function requires more arguments, consider using the
            'LossFunctionWrapper' class from 'orqviz.loss_function'.
        direction: Direction in which loss function is scanned around the origin.
            If None, a random unit vector is sampled. Defaults to None
        n_points: Number of points to evaluate along the scan. Defaults to 31.
        end_points: Range of scan along the direction in units of the direction vector.
            Defaults to (-np.pi, np.pi).
        verbose: Flag for printing progress. Defaults to False.
    """
    if direction is None:
        direction = get_random_normal_vector(np.shape(origin))

    point_list: ArrayOfParameterVectors = direction_linspace(
        origin=origin,
        direction=direction,
        n_points=n_steps,
        endpoints=end_points,
    )

    scan_values = eval_points_on_path(point_list, loss_function, verbose=verbose)
    return Scan1DResult(
        point_list, direction=direction, values=scan_values, origin=origin
    )


def perform_1D_interpolation(
    point_1: ParameterVector,
    point_2: ParameterVector,
    loss_function: LossFunction,
    n_steps: int = 100,
    end_points: Tuple[float, float] = (-0.5, 1.5),
    parameter_period: Optional[float] = None,
    verbose: bool = False,
) -> Scan1DResult:
    """Function to perform a 1D scan to interpolate between two points.

    Args:
        point_1: First point of the interpolation.
        point_2: Second point of the interpolation.
        loss_function: Function to perform the scan on. It must receive only a
            numpy.ndarray of parameters, and return a real number.
            If your function requires more arguments, consider using the
            'LossFunctionWrapper' class from 'orqviz.loss_function'.
        n_steps: Number of points evaluated along the scan. Defaults to 100.
        end_points: Range of scan along the direction in units of the
            interpolation vector. Defaults to (-0.5, 1.5).
        parameter_period: Optional period of the parameters to scan the shortest
            interpolated path between the points.
            If None, interpolation per parameter happens along the real number line.
            Defaults to None.
        verbose: Flag for printing progress. Defaults to False.
    """
    if parameter_period:
        point_2 = relative_periodic_wrap(point_1, point_2, parameter_period)

    origin = point_1
    direction = point_2 - point_1
    point_list: ArrayOfParameterVectors = direction_linspace(
        origin=origin,
        direction=direction,
        n_points=n_steps,
        endpoints=end_points,
    )

    loss_vector = eval_points_on_path(
        point_list, loss_function=loss_function, verbose=verbose
    )

    return Scan1DResult(
        point_list,
        direction=direction,
        values=loss_vector,
        origin=origin,
    )



import time
from typing import Callable, Optional

import numpy as np


def _calculate_new_average(
    previous_average: Optional[float], count: int, new_value: float
) -> float:
    if previous_average is None:
        return new_value
    else:
        return (count * previous_average + new_value) / (count + 1)


class LossFunctionWrapper:
    def __init__(self, loss_function: Callable, *args, **kwargs):
        def wrapped_loss_function(params: np.ndarray) -> float:
            return loss_function(params, *args, **kwargs)

        self.loss_function = wrapped_loss_function
        self.call_count = 0
        self.average_call_time: Optional[float] = None
        self.min_value: Optional[float] = None

    def __call__(self, params: np.ndarray) -> float:
        start_time = time.perf_counter()
        value = self.loss_function(params)
        total_time = time.perf_counter() - start_time
        self.average_call_time = _calculate_new_average(
            self.average_call_time, self.call_count, total_time
        )
        self.call_count += 1
        if self.min_value is None or value < self.min_value:
            self.min_value = value
        return value

    def reset(self) -> None:
        self.call_count = 0
        self.average_call_time = None
        self.min_value = None



import warnings
from typing import Optional, Tuple

import matplotlib.pyplot as plt
import matplotlib.ticker as tck
import mpl_toolkits
import numpy as np
from matplotlib.cm import ScalarMappable


def normalize_color_and_colorbar(
    fig: Optional[plt.Figure] = None,
    ax: Optional[plt.Axes] = None,
    min_val: float = 0.0,
    max_val: float = 1.0,
    cmap: str = "viridis",
    image_index: int = 0,
):
    """Function to adjust the color and colorbar of a matplotlib plot.

    Args:
        fig: Matplotlib figure in which the plot is performed.
            If None, gets current figure. Defaults to None.
        ax: Matplotlib axis in which the plot is performed.
            If None, gets current axis. Defaults to None.
        min_val: Minimum values of the image and colorbar range. Defaults to 0.0.
        max_val: Maximum values of the image and colorbar range. Defaults to 1.0.
        cmap: Matplotlib colormap for the plot. Defaults to "viridis".
        image_index: Position index for the image in the Matplotlib axis.
            Defaults to 0.

    """
    fig, ax = _check_and_create_fig_ax(fig, ax)

    try:
        image = ax.collections[image_index]
    except (AttributeError, IndexError):
        try:
            image = ax.images[image_index]
        except (AttributeError, IndexError) as e:
            raise ValueError(
                "Provided ax does not contain an image in ax.images or ax.collections"
            ) from e

    assert isinstance(image, ScalarMappable)
    if image.colorbar is not None:
        image.colorbar.remove()
    image.set_clim(vmin=min_val, vmax=max_val)  # type: ignore
    image.set_cmap(cmap)  # type: ignore
    fig.colorbar(image, ax=ax)


def get_colorbar_from_ax(
    ax: Optional[plt.Axes] = None,
    image_index: Optional[int] = None,
):
    """Helper function to extract the colorbar of a previously created plot.

    Args:
        ax: Matplotlib axis in which the colorbar was created.
            If None, gets current axis. Defaults to None.
        image_index: Position index for the image in the Matplotlib axis.
            If None, will return the first colorbar that is found. Defaults to None.

    """
    _, ax = _check_and_create_fig_ax(ax=ax)

    if image_index is None:
        len_collections = len(ax.collections)
        len_images = len(ax.images)

        if len_collections > 0:
            for ii in range(len_collections):
                image = ax.collections[ii]
                if hasattr(image, "colorbar"):
                    return image.colorbar

        elif len_images > 0:
            for ii in range(len_images):
                image = ax.images[ii]
                if hasattr(image, "colorbar"):
                    return image.colorbar

        raise AttributeError(
            "Provided ax does not contain an image with a colorbar in ax.images"
            " or ax.collections"
        )

    else:
        try:
            image = ax.collections[image_index]
        except (AttributeError, IndexError):
            try:
                image = ax.images[image_index]
            except (AttributeError, IndexError) as e:
                raise ValueError(
                    "Provided ax does not contain an image "
                    "in ax.images or ax.collections"
                ) from e


def set_ticks_to_multiples_of_pi(
    ax: Optional[plt.Axes] = None,
    base: float = np.pi / 2,
):
    """Helper function to set the ticks of matplotlib axes
    to multiples of pi with pi symbols."""
    _, ax = _check_and_create_fig_ax(ax=ax)

    ax.xaxis.set_major_formatter(
        tck.FuncFormatter(
            lambda val, pos: "{:.2f}$\pi$".format(val / np.pi)  # noqa: W605
            if val != 0
            else "0"
        )
    )
    ax.xaxis.set_major_locator(tck.MultipleLocator(base=base))

    ax.yaxis.set_major_formatter(
        tck.FuncFormatter(
            lambda val, pos: "{:.2f}$\pi$".format(val / np.pi)  # noqa: W605
            if val != 0
            else "0"
        )
    )
    ax.yaxis.set_major_locator(tck.MultipleLocator(base=base))


def set_ticks_to_multiples_of_tau(
    ax: Optional[plt.Axes] = None,
    base: float = np.pi,
):
    """Helper function to set the ticks of matplotlib axes
    to multiples of tau with tau symbols."""
    _, ax = _check_and_create_fig_ax(ax=ax)

    tau = 2 * np.pi
    ax.xaxis.set_major_formatter(
        tck.FuncFormatter(
            lambda val, pos: "{:.2f}$\\tau$".format(val / tau)  # noqa: W605
            if val != 0
            else "0"
        )
    )
    ax.xaxis.set_major_locator(tck.MultipleLocator(base=base))

    ax.yaxis.set_major_formatter(
        tck.FuncFormatter(
            lambda val, pos: "{:.2f}$\\tau$".format(val / tau)  # noqa: W605
            if val != 0
            else "0"
        )
    )
    ax.yaxis.set_major_locator(tck.MultipleLocator(base=base))


def _check_and_create_fig_ax(
    fig: Optional[plt.Figure] = None,
    ax: Optional[plt.Axes] = None,
) -> Tuple[plt.Figure, plt.Axes]:
    if fig is None:
        fig = plt.gcf()

    if ax is None:
        ax = fig.gca()

    return fig, ax


def _check_and_create_3D_ax(
    ax: Optional[plt.Axes] = None,
) -> mpl_toolkits.mplot3d.Axes3D:
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(projection="3d")
    elif ax.name != "3d":
        warnings.warn(
            "The matplotlib axis you provided is not a 3d axis. "
            "Your axis is overridden with a new axis."
        )
        warnings.warn("You can create a 3d axis with fig.add_subplot(projection='3d')")
        fig = plt.figure()
        ax = fig.add_subplot(projection="3d")

    return ax



from typing import Callable, Optional, Tuple

import numpy as np

from orqviz.aliases import (
    ArrayOfParameterVectors,
    FullGradientFunction,
    LossFunction,
    ParameterVector,
)
from orqviz.gradients import calculate_full_gradient


def gradient_descent_optimizer(
    init_params: ParameterVector,
    loss_function: LossFunction,
    n_iters: int,
    learning_rate: float = 0.1,
    full_gradient_function: FullGradientFunction = None,
    eval_loss_during_training: bool = True,
) -> Tuple[ArrayOfParameterVectors, np.ndarray]:
    """Function perform gradient descent optimization on a loss function.

    Args:
        init_params: Initial parameter vector from which to start the optimization.
        loss_function: Function with respect to which the gradient is calculated.
            It must receive only a numpy.ndarray of parameters, and return
            a real number. If your function requires more arguments, consider using the
            'LossFunctionWrapper' class from 'orqviz.loss_function'.
        n_iters: Number of iterations to optimize.
        learning_rate: Learning rate for gradient descent. The calculated gradient
            is multiplied with this value and then updates the parameter vector.
        full_gradient_function: Gradient function to calculate the partial derivatives
            of the loss function with respect to each parameter vector entry.
            If None, a simple numerical gradient is computed. Defaults to None.
        eval_loss_during_training: Flag to indicate whether to evaluate the loss
            at every iteration during training. Doing so adds 'n_iters-1'
            loss function calls. If False, only the initial and final losses
            are calculated and returned. Defaults to False.
    """
    if full_gradient_function is None:

        def _full_gradient_function(params):
            return calculate_full_gradient(params, loss_function=loss_function)

        full_gradient_function = _full_gradient_function

    params = init_params
    all_costs = []
    all_params = [init_params]
    for _ in range(n_iters):
        if eval_loss_during_training:
            all_costs.append(loss_function(params))
        grad_val = full_gradient_function(params)
        params = params - learning_rate * grad_val
        all_params.append(params)

    all_costs.append(loss_function(params))

    return np.array(all_params), np.array(all_costs)



from typing import Callable, List, Optional

import numpy as np

from ..aliases import (
    DirectionVector,
    FullGradientFunction,
    LossFunction,
    ParameterVector,
    Weights,
)
from ..gradients import calculate_full_gradient
from .data_structures import Chain, ChainPath


def run_NEB(
    init_chain: Chain,
    loss_function: LossFunction,
    full_gradient_function: Optional[FullGradientFunction] = None,
    n_iters: int = 10,
    eps: float = 1e-3,
    learning_rate: float = 0.1,
    stochastic: bool = False,
    calibrate_tangential: bool = False,
    cummulative_weights: Optional[Weights] = None,
    verbose: bool = False,
) -> List[Chain]:
    """Nudged Elastic Band (NEB) algorithm to train a piece-wise linear path optimized
        in the loss function landscape.
    Reference paper:
        Essentially No Barriers in Neural Network Energy Landscape, arXiv:1803.00885.
    NOTE: There is a discrepancy between name used in the original paper (elastic band)
        and the one we use to define our data structure (chain), but we decided to use
        the later as we feel it's more appropriate for this data structure.

    Args:
        init_chain: Initial chain that is optimized with the algorithm.
        loss_function: Function that is used to optimize the chain. It must
            receive only a numpy.ndarray of parameters, and return a real number.
            If your function requires more arguments, consider using the
            'LossFunctionWrapper' class from 'orqviz.loss_function'.
        full_gradient_function: Function to calculate the gradient w.r.t.
            the loss function for all parameters. Defaults to None.
        n_iters: Number of optimization iterations. Defaults to 10.
        eps: Stencil for finite difference gradient if full_gradient_function
            is not provided. For noisy loss functions,
            we recommend increasing this value. Defaults to 1e-3.
        learning_rate: Learning rate/ step size for the gradient descent optimization.
            Defaults to 0.1.
        stochastic: Flag to indicate whether to perform stochastic gradient descent
            if full_gradient_function is not provided.
            It is less stable but much faster. Defaults to False.
        calibrate_tangential: Flag to indicate whether next neighbor for finding
            tangential direction is calibrated with an additional loss evaluation.
            Defaults to False.
        cummulative_weights: Cummulative chain position weights to re-distributed
            pivots along chain. If None, pivots are re-distributed uniformly.
            Defaults to None.
        verbose: Flag for printing progress. Defaults to False.

    Returns:
        All chains during training
    """
    if full_gradient_function is None:
        # Defines an automatic numerical gradient with eps, stochastic as args
        def _full_gradient_function(pars: ParameterVector) -> ParameterVector:
            return calculate_full_gradient(
                pars, loss_function=loss_function, stochastic=stochastic, eps=eps
            )

        full_gradient_function = _full_gradient_function

    all_chains = [init_chain]
    chain = init_chain
    for it in range(n_iters):
        if it % 5 == 0 and verbose:
            print("iteration", it)
        gradients_on_pivots = _get_gradients_on_pivots(
            chain,
            loss_function,
            full_gradient_function,
            calibrate_tangential,
        )
        new_pivots = np.asarray(
            [
                old_pivot - learning_rate * gradient
                for old_pivot, gradient in zip(chain.pivots, gradients_on_pivots)
            ]
        )
        chain = Chain(new_pivots)
        chain = _redistribute_chain(chain, cummulative_weights)
        all_chains.append(chain)

    return all_chains


def _get_gradients_on_pivots(
    chain: Chain,
    loss_function: LossFunction,
    full_gradient_function: FullGradientFunction,
    calibrate_tangential: bool = False,
) -> np.ndarray:
    """Calculates gradient for every pivot on the chain w.r.t. the loss function
        using the gradient function.

    Args:
        chain: Chain to calculate the gradients on.
        loss_function: Function that is used to optimize the chain. It must receive
            only a numpy.ndarray of parameters, and return a real number.
            If your function requires more arguments, consider using the
            'LossFunctionWrapper' class from 'orqviz.loss_function'.
        full_gradient_function: Function to calculate the gradient w.r.t.
            the loss function for all parameters.
        calibrate_tangential: Flag to indicate whether next neighbor for finding
            tangential direction is calibrated with an additional loss evaluation.
            Defaults to False.
    """

    # We initialize with zeros, as we always want first and last gradient
    # to be equal to 0.
    gradients_on_pivots = np.zeros(shape=(chain.n_pivots, *chain.param_shape))

    for ii in range(1, chain.n_pivots - 1):
        before = chain.pivots[ii - 1]
        this = chain.pivots[ii]
        after = chain.pivots[ii + 1]
        #
        full_grad = full_gradient_function(this)
        #
        tan = this - before
        if calibrate_tangential and loss_function(after) > loss_function(before):
            tan = after - this
        tan /= np.linalg.norm(tan)
        ax_indices = tuple(range(len(full_grad.shape)))
        tangential_grad = (
            np.tensordot(full_grad, tan, axes=(ax_indices, ax_indices)) * tan
        )
        # save update
        gradients_on_pivots[ii] = full_grad - tangential_grad

    return gradients_on_pivots


def _redistribute_chain(
    chain: Chain,
    cummulative_weights: Optional[Weights] = None,
) -> Chain:
    """Helper function to re-distribute pivots along a Chain according
    to their weights on the chain."""
    path = ChainPath(chain)
    if cummulative_weights is not None:
        return path._get_chain_from_weights(cummulative_weights)
    else:
        return path.generate_uniform_chain(chain.n_pivots)



from typing import Optional

import matplotlib
import numpy as np

from .aliases import ArrayOfParameterVectors
from .geometric import get_coordinates_on_direction
from .plot_utils import _check_and_create_fig_ax


def plot_optimization_trajectory(
    optimization_trajectory: ArrayOfParameterVectors,
    direction_x: np.ndarray,
    direction_y: np.ndarray,
    ax: matplotlib.axes.Axes,
    shift: Optional[np.ndarray] = None,
    **plot_kwargs,
):
    """Function to project and plot a parameter trajectory on a 2D plane.

    Args:
        optimization_trajectory: Parameter trajectory to be projected
            and plotted on a 2D plane.
        direction_x: x-direction of the 2D plane.
        direction_y: y-direction of the 2D plane.
        ax: Matplotlib axis to perform plot on. If None, a new axis
            is created from the current figure. Defaults to None.
        shift: Origin to shift the trajectory to. Defaults to None.
        plot_kwargs: kwargs for plotting with matplotlib.pyplot.plot (plt.plot)
    """

    _, ax = _check_and_create_fig_ax(ax=ax)

    projected_trajectory_x = get_coordinates_on_direction(
        optimization_trajectory, direction_x, origin=shift
    )
    projected_trajectory_y = get_coordinates_on_direction(
        optimization_trajectory, direction_y, origin=shift
    )

    default_plot_kwargs = {
        "color": "lightgray",
        "linestyle": "-",
        "marker": ".",
        "alpha": 0.8,
    }
    plot_kwargs = {**default_plot_kwargs, **plot_kwargs}

    ax.plot(
        projected_trajectory_x,
        projected_trajectory_y,
        **plot_kwargs,
    )
    ax.plot(
        projected_trajectory_x[0],
        projected_trajectory_y[0],
        c=plot_kwargs["color"],
        marker="s",
        alpha=plot_kwargs["alpha"],
    )
    ax.plot(
        projected_trajectory_x[-1],
        projected_trajectory_y[-1],
        c=plot_kwargs["color"],
        marker="*",
        alpha=plot_kwargs["alpha"],
    )


def plot_scatter_points(
    scatter_points: ArrayOfParameterVectors,
    direction_x: np.ndarray,
    direction_y: np.ndarray,
    ax: Optional[matplotlib.axes.Axes] = None,
    shift: Optional[np.ndarray] = None,
    **plot_kwargs,
):
    """Function to project and scatter plot an array of points on a 2D plane.

    Args:
        scatter_points: Points to be to be projected and scattered on a 2D plane.
        direction_x: x-direction of the 2D plane.
        direction_y: y-direction of the 2D plane.
        ax: Matplotlib axis to perform plot on. If None, a new axis
            is created from the current figure. Defaults to None.
        shift: Origin to shift the trajectory to. Defaults to None.
        plot_kwargs: kwargs for plotting with matplotlib.pyplot.scatter (plt.scatter)
    """

    _, ax = _check_and_create_fig_ax(ax=ax)

    projected_scatter_x = get_coordinates_on_direction(
        scatter_points, direction_x, origin=shift
    )
    projected_scatter_y = get_coordinates_on_direction(
        scatter_points, direction_y, origin=shift
    )

    default_plot_kwargs = {"color": "lightgray", "marker": ".", "alpha": 0.8}
    plot_kwargs = {**default_plot_kwargs, **plot_kwargs}

    ax.scatter(
        projected_scatter_x,
        projected_scatter_y,
        **plot_kwargs,
    )


def plot_line_through_points(
    points: ArrayOfParameterVectors,
    direction_x: np.ndarray,
    direction_y: np.ndarray,
    ax: Optional[matplotlib.axes.Axes] = None,
    shift: Optional[np.ndarray] = None,
    **plot_kwargs,
):
    """Function to project and points on a 2D plane and plot a line through them.

    Args:
        points: Points to be to be projected on a 2D plane and plot a line through.
        direction_x: x-direction of the 2D plane.
        direction_y: y-direction of the 2D plane.
        ax: Matplotlib axis to perform plot on. If None, a new axis is created
            from the current figure. Defaults to None.
        shift: Origin to shift the trajectory to. Defaults to None.
        plot_kwargs: kwargs for plotting with matplotlib.pyplot.plot (plt.plot)
    """

    _, ax = _check_and_create_fig_ax(ax=ax)

    projected_scatter_x = get_coordinates_on_direction(
        points, direction_x, origin=shift
    )
    projected_scatter_y = get_coordinates_on_direction(
        points, direction_y, origin=shift
    )

    default_plot_kwargs = {"color": "lightgray", "alpha": 0.8}
    plot_kwargs = {**default_plot_kwargs, **plot_kwargs}

    ax.plot(
        projected_scatter_x,
        projected_scatter_y,
        **plot_kwargs,
    )



from typing import Callable, List, Optional

import numpy as np
from scipy.interpolate import interp1d

from ..aliases import FullGradientFunction, LossFunction, ParameterVector
from .data_structures import Chain
from .neb import run_NEB


# Nudged-Elastic-Band
def run_AutoNEB(
    init_chain: Chain,
    loss_function: LossFunction,
    full_gradient_function: Optional[FullGradientFunction] = None,
    n_cycles: int = 4,
    n_iters_per_cycle: int = 10,
    max_new_pivots: int = 1,
    weighted_redistribution: bool = False,
    insert_at_beginning: bool = False,
    percentage_tol: float = 0.2,
    absolute_tol: float = 0.0,
    eps: float = 0.1,
    learning_rate: float = 0.1,
    stochastic: bool = False,
    calibrate_tangential: bool = False,
    verbose: bool = False,
) -> List[Chain]:
    """Automatic wrapping of the NEB algorithm that dynamically inserts pivots
    to the piece-wise linear path being optimized in the loss function landscape.
    Returns a list of all chains during training where index -1
    is the final optimized chain.
    Reference paper:
        Essentially No Barriers in Neural Network Energy Landscape, arXiv:1803.00885.
    NOTE:
        There is a discrepancy between name used in the original paper (elastic band)
        and the one we use to define our data structure (chain), but we decided
        to use the later as we feel it's more appropriate for this data structure.

    Args:
        init_chain: Initial chain that is optimized with the algorithm.
        loss_function: Function that is used to optimize the chain. It must receive
            only a numpy.ndarray of parameters, and return a real number.
            If your function requires more arguments, consider using the
            'LossFunctionWrapper' class from 'orqviz.loss_function'.
        full_gradient_function: Function to calculate the gradient w.r.t.
            the loss function for all parameters. Defaults to None.
        n_cycles: Number of cycles between which new pivots can be inserted.
            Defaults to 4.
        n_iters_per_cycle: Number of optimization iterations per cycle.
            Defaults to 10.
        max_new_pivots: Maximum number of pivots inserted per cycle. Defaults to 1.
        weighted_redistribution: Flag to indicate whether pivots are uniformly
            re-distributed along the chain or according to their insertion position
            in the chain. Defaults to False.
        insert_at_beginning: Flag to indicate whether to insert pivots
            before first cycle. Defaults to False.
        percentage_tol: Percentage error threshold to insert new pivots.
            Called 'alpha' in the original paper. Be mindful of the magnitude and
            sign of typical loss values. Additive to absolute error. Defaults to 0.2.
        absolute_tol: Absolute error threshold to insert new pivots.
            Additive to percentage error. Defaults to 0.0.
        eps: Stencil for finite difference gradient if gradient_function
            is not provided. Defaults to 0.1.
        learning_rate: Learning rate/ step size for the gradient
            descent optimization. Defaults to 0.1.
        stochastic: Flag to indicate whether to perform stochastic gradient descent
            if gradient_function is not provided. It is less stable but much faster.
            Defaults to False.
        calibrate_tangential: Flag to indicate whether next neighbor for finding
            tangential direction is calibrated with an additional loss evaluation.
            Defaults to False.
        verbose: Flag for printing progress. Defaults to False.

    Returns:
        All chains during training
    """
    chain = init_chain

    all_chains = [chain]
    if insert_at_beginning:
        # find new pivots
        chain = _insert_pivots_to_improve_approximation(
            chain=chain,
            loss_function=loss_function,
            max_new_pivots=max_new_pivots,
            percentage_tol=percentage_tol,
            absolute_tol=absolute_tol,
        )

    for cycle in range(n_cycles):
        cummulative_weights = chain.get_weights() if weighted_redistribution else None
        optimization_history = run_NEB(
            init_chain=chain,
            loss_function=loss_function,
            full_gradient_function=full_gradient_function,
            n_iters=n_iters_per_cycle,
            cummulative_weights=cummulative_weights,
            eps=eps,
            learning_rate=learning_rate,
            stochastic=stochastic,
            calibrate_tangential=calibrate_tangential,
            verbose=verbose,
        )
        chain = optimization_history[-1]
        all_chains += optimization_history

        if cycle != n_cycles - 1:
            # find new pivots
            chain = _insert_pivots_to_improve_approximation(
                chain=chain,
                loss_function=loss_function,
                max_new_pivots=max_new_pivots,
                percentage_tol=percentage_tol,
                absolute_tol=absolute_tol,
            )
            all_chains.append(chain)

    return all_chains


def _insert_pivots_to_improve_approximation(
    chain: Chain,
    loss_function: LossFunction,
    max_new_pivots: int = 1,
    percentage_tol: float = 0.2,
    absolute_tol: float = 0.0,
) -> Chain:
    """Method to check where the piece-wise linear approximation of the current Chain
        with respect to the loss function is not good enough. If that's the case,
        it adds new pivots to the chain.

    Args:
        chain: Current Chain
        loss_function: Function for NEB training. It must receive only a
            numpy.ndarray of parameters, and return a real number.
            If your function requires more arguments, consider using the
            'LossFunctionWrapper' class from 'orqviz.loss_function'.
        max_new_pivots: Maximum number of pivots inserted to Chain. Defaults to 1.
        percentage_tol: Percentage error threshold to insert new pivots.
            Be mindful of the magnitude and sign of typical loss values.
            Additive to absolute error. Defaults to 0.2.
        absolute_tol: Absolute error threshold to insert new pivots.
            Additive to percentage error. Defaults to 0.0.
    """

    chain_weights = chain.get_weights()

    losses = chain.evaluate_on_pivots(loss_function)
    linear_weight_loss_map = interp1d(chain_weights, losses, kind="linear")
    linear_weight_chain_map = interp1d(
        chain_weights, chain.pivots, kind="linear", axis=0
    )
    # Check between each pair of pivots if an additional pivot
    # would improve the approximation
    trial_cum_weights = [
        (chain_weights[ii + 1] + chain_weights[ii]) / 2
        for ii in range(len(chain_weights) - 1)
    ]

    trial_points = linear_weight_chain_map(trial_cum_weights)
    trial_losses = np.array([loss_function(point) for point in trial_points])
    interpolated_losses = linear_weight_loss_map(trial_cum_weights)

    residuals = trial_losses - interpolated_losses
    significant_residuals = np.where(
        np.abs(residuals) > percentage_tol * interpolated_losses + absolute_tol,
        residuals,
        0,
    )
    new_pivots = chain.pivots

    if not np.isclose(significant_residuals, 0).all():
        largest_residual_positions = np.argsort(significant_residuals)[::-1]
        # sorts the most important indices backwards to insert correctly into the chain
        indices_to_select = sorted(
            largest_residual_positions[:max_new_pivots], reverse=True
        )
        for piv in indices_to_select:
            if np.isclose(trial_losses[piv], 0):
                continue
            new_cum_weight = trial_cum_weights[piv]
            insert_index = next(
                k for k, w in enumerate(chain_weights) if w > new_cum_weight
            )
            new_pivots = np.insert(
                new_pivots,
                insert_index,
                linear_weight_chain_map(new_cum_weight),
                axis=0,
            )

    return Chain(new_pivots)



from typing import Optional, Tuple, Union

import numpy as np
from scipy.interpolate import interp1d

from .aliases import ArrayOfParameterVectors, DirectionVector, ParameterVector


def get_random_normal_vector(dimension: Union[int, Tuple]) -> DirectionVector:
    """Helper function to generate a vector with a specified dimension and norm=1."""
    random_vector = np.random.normal(0, 1, size=dimension)
    return random_vector / np.linalg.norm(random_vector)


def get_random_orthonormal_vector(base_vector: DirectionVector) -> DirectionVector:
    """Helper function to generate a random orthogonal vector with respect to
    a provided base vector."""
    random_vector = np.random.normal(size=np.shape(base_vector))
    new_vector = (
        random_vector
        - np.dot(random_vector.flatten(), base_vector.flatten())
        * base_vector
        / np.linalg.norm(base_vector) ** 2
    )
    return new_vector / np.linalg.norm(new_vector)


def relative_periodic_wrap(
    reference_point: ParameterVector,
    point_to_be_wrapped: ParameterVector,
    period: float = 2 * np.pi,
) -> ParameterVector:
    """Function that returns a wrapped 'copy' of a point to be wrapped such that
        the distance between it and the reference point is minimal inside
        the specified period.

    Args:
        reference_point: Reference point for periodic wrapping of the second point.
        point_to_be_wrapped: Point that is wrapped to a copy of itself such that
            its distance to the reference point is minimal inside the specified period.
        period: Periodicity of each parameter of the the point that is to be wrapped.
            Defaults to 2*np.pi.
    """
    option1 = (reference_point - point_to_be_wrapped) % period
    option2 = (point_to_be_wrapped - reference_point) % period
    diff = np.where(option1 > option2, option2, -option1)
    wrapped_point = reference_point + diff
    return wrapped_point


def relative_periodic_trajectory_wrap(
    reference_point: ParameterVector,
    trajectory: ArrayOfParameterVectors,
    period: float = 2 * np.pi,
) -> ArrayOfParameterVectors:
    """Function that returns a wrapped 'copy' of a parameter trajectory such that
        the distance between the final point of the trajectory and the reference point
        is minimal inside the specified period.
        The rest of the trajectory is being transformed in the same manner.
        NOTE:
            It only works as intended if the period is larger than the distance
            between the consecutive points in the trajectory.

    Args:
        reference_point: Reference point for periodic wrapping of the trajectory.
        trajectory: Trajectory that is wrapped to a copy of itself such that
            the distance between the final point in the trajectory
            and the reference point is minimal.
        period: Periodicity of each parameter in each point of the trajectory.
            Defaults to 2*np.pi.
    """
    if not np.all(np.linalg.norm(np.diff(trajectory, axis=0), axis=1) < period):
        raise ValueError(
            "Distances between consecutive points must be smaller than period."
        )
    wrapped_trajectory = np.copy(trajectory).astype(float)
    wrapped_trajectory[-1] = relative_periodic_wrap(
        reference_point, trajectory[-1], period=period
    )
    for ii in range(2, len(wrapped_trajectory) + 1):
        wrapped_trajectory[-ii] = relative_periodic_wrap(
            wrapped_trajectory[-ii + 1], trajectory[-ii], period=period
        )

    return wrapped_trajectory


def get_coordinates_on_direction(
    points: ArrayOfParameterVectors,
    direction: DirectionVector,
    origin: Optional[ParameterVector] = None,
    in_units_of_direction: bool = False,
) -> np.ndarray:
    """Helper function to calculate the projection of points onto a direction
        to extract the coordinates in that direction.

    Args:
        points: Points to be projected on the direction vector.
        direction: Direction vector to project the points on.
        origin: Origin for the projection. Defaults to None.
        in_units_of_direction: Flag to indicate whether to return coordinates in units
            of the direction vector.
            If False, returns coordinates in euclidean distances. Defaults to False.
    """
    if origin is not None:
        points = points - origin
    norm_direction = np.linalg.norm(direction)
    if in_units_of_direction:
        direction = direction / norm_direction
    return (
        np.tensordot(
            points,
            direction,
            axes=(range(1, len(points.shape)), range(len(direction.shape))),
        )
        / norm_direction
    )


def direction_linspace(
    origin: ParameterVector,
    direction: DirectionVector,
    n_points: int,
    endpoints: Tuple[float, float] = (-1, 1),
) -> ArrayOfParameterVectors:
    """Helper function to wrap np.linspace in order to create points on a specified
    direction around an origin."""
    return np.linspace(
        origin + endpoints[0] * direction,
        origin + endpoints[1] * direction,
        num=n_points,
    )


def uniformly_distribute_trajectory(
    parameter_trajectory: ArrayOfParameterVectors,
    n_points: int,
) -> ArrayOfParameterVectors:
    """Function to distribute points uniformly (in euclidean distance) along a path
    that is given by a parameter trajectory, i.e. an array of parameter vectors.
    Returns an array of parameter vectors where the first and last entries match
    those of the passed parameter trajectory and all entries are equally distant
    in euclidean distance.
    """
    trajectory_weights = np.linalg.norm(np.diff(parameter_trajectory, axis=0), axis=1)
    trajectory_weights /= np.sum(trajectory_weights)
    cum_weights = np.cumsum(trajectory_weights)
    cum_weights = np.insert(cum_weights, 0, 0)
    cum_weights[-1] = 1  # for stability insert an integer 1

    weight_interpolator = interp1d(
        cum_weights,
        parameter_trajectory,
        axis=0,
    )
    eval_points = np.linspace(0, 1, num=n_points)
    return weight_interpolator(eval_points)


def _norm_of_arrayofparametervectors(param_array: ArrayOfParameterVectors):
    ax_indices = tuple(range(len(param_array.shape)))
    t_dot = np.tensordot(
        param_array, param_array, axes=(ax_indices[1:], ax_indices[1:])
    )
    return np.array(np.sqrt(np.diag(t_dot)))



from dataclasses import asdict, dataclass, field
from typing import List, Optional, Tuple

import numpy as np

from ..aliases import DirectionVector, ParameterVector


@dataclass()
class HessianEigenobject:
    """Data structure for Hessian matrix.
        Eigenvalues and Eigenvectors are automatically calculated and sorted.

    Args:
        params: Parameter vector at which the Hessian matrix was measured
        hessian_matrix: Hessian matrix evaluated at params as a 2D numpy array.
    """

    params: ParameterVector
    hessian_matrix: np.ndarray
    eigenvectors: List[DirectionVector] = field(init=False)
    eigenvalues: np.ndarray = field(init=False)

    def __post_init__(self):
        eigenvalues, eigenvectors = np.linalg.eig(self.hessian_matrix)
        eigenvectors = eigenvectors.T
        sorted_ind = np.argsort(eigenvalues)
        self.eigenvalues = eigenvalues[sorted_ind]
        self.eigenvectors = eigenvectors[sorted_ind].reshape((-1, *self.params.shape))



from typing import Callable

import numpy as np

"""
ParameterVector, ArrayOfParameterVectors, GridOfParameterVectors
are all aliases of numpy.ndarray and are used throughout this library.
They indicate whether the numpy.ndarray that is expected by a method or class
is a 1D, 2D, or 3D numpy.ndarray, respectively. The last dimension, for ParameterVector
the only dimension, is always of size number_of_parameters, while the other dimensions
indicate how many of them there are.
"""
ParameterVector = np.ndarray  # ND array
ArrayOfParameterVectors = np.ndarray  # Array of ND arrays
GridOfParameterVectors = np.ndarray  # Grid of ND arrays
Weights = np.ndarray  # 1D vector of floats from 0-1
DirectionVector = np.ndarray  # ND array with same shape as ParameterVector
LossFunction = Callable[
    [ParameterVector], float
]  # Function that can be scanned with orqviz
GradientFunction = Callable[
    [ParameterVector, DirectionVector], float
]  # Returns partial derrivative of LossFunction wrt DirectionVector
FullGradientFunction = Callable[
    [ParameterVector], np.ndarray
]  # Returns all partial derrivatives of LossFunction wrt each parameter



from .data_structures import HessianEigenobject
from .hessians import (
    get_Hessian,
    get_Hessian_SPSA_approx,
    perform_1D_hessian_eigenvector_scan,
)
from .plots import plot_1D_hessian_eigenvector_scan_result



from typing import NamedTuple, Optional

import numpy as np

from ..aliases import ArrayOfParameterVectors, GridOfParameterVectors, ParameterVector
from ..geometric import get_coordinates_on_direction


class Scan1DResult(NamedTuple):
    """Datatype for 1D scans to combine the scan result and scan instruction."""

    params_list: ArrayOfParameterVectors
    direction: np.ndarray
    values: np.ndarray
    origin: Optional[ParameterVector] = None

    def _get_coordinates_on_direction(self, in_units_of_direction: bool = False):
        """Projects parameters in params_vector and projects them on the direction_x."""
        return get_coordinates_on_direction(
            self.params_list,
            self.direction,
            origin=self.origin,
            in_units_of_direction=in_units_of_direction,
        )


class Scan2DResult(NamedTuple):
    """Datatype for 2D scans to combine the scan result and scan instruction."""

    params_grid: GridOfParameterVectors
    direction_x: np.ndarray
    direction_y: np.ndarray
    values: np.ndarray
    origin: Optional[ParameterVector] = None

    def _get_coordinates_on_directions(self, in_units_of_direction: bool = False):
        """Projects parameters in directions x and y."""
        x_values = get_coordinates_on_direction(
            self.params_grid[0, :],
            self.direction_x,
            origin=self.origin,
            in_units_of_direction=in_units_of_direction,
        )
        y_values = get_coordinates_on_direction(
            self.params_grid[:, 0],
            self.direction_y,
            origin=self.origin,
            in_units_of_direction=in_units_of_direction,
        )
        return x_values, y_values


def clone_Scan1DResult_with_different_values(
    base_scan: Scan1DResult, values: np.ndarray
) -> Scan1DResult:
    """Helper function that returns a copy of a Scan1DResult object with new values"""
    class_dict = base_scan._asdict()
    class_dict["values"] = values
    return Scan1DResult(**class_dict)


def clone_Scan2DResult_with_different_values(
    base_scan: Scan2DResult, values: np.ndarray
) -> Scan2DResult:
    """Helper function that returns a copy of a Scan2DResult object with new values"""
    class_dict = base_scan._asdict()
    class_dict["values"] = values
    return Scan2DResult(**class_dict)



import warnings
from typing import Union

from .elastic_band.data_structures import Chain
from .hessians.data_structures import HessianEigenobject
from .io import load_viz_object as _load_viz_object
from .io import save_viz_object as _save_viz_object
from .scans.data_structures import Scan1DResult, Scan2DResult

OrqVizObject = Union[Scan1DResult, Scan2DResult, HessianEigenobject, Chain]


def save_viz_object(viz_object: OrqVizObject, filename: str) -> None:
    """Save datatype to a with pickle"""
    warnings.warn(
        """orqviz.utils.save_viz_object is deprecated,
        please use orqviz.io.save_viz_object""",
        DeprecationWarning,
    )
    _save_viz_object(viz_object, filename)


def load_viz_object(filename: str) -> OrqVizObject:
    """Load datatype from a file with pickle"""
    warnings.warn(
        """orqviz.utils.load_viz_object is deprecated,
        please use orqviz.io.load_viz_object""",
        DeprecationWarning,
    )
    return _load_viz_object(filename)



from typing import Callable, List, Optional

import numpy as np

from ..aliases import (
    ArrayOfParameterVectors,
    GridOfParameterVectors,
    LossFunction,
    ParameterVector,
)


def eval_points_on_path(
    all_points: ArrayOfParameterVectors,
    loss_function: LossFunction,
    n_reps: int = 1,
    verbose: bool = False,
) -> np.ndarray:
    """Function to evaluate loss function on a 1D path of parameters.

    Args:
        all_parameters: Array of parameters with shape (len, *(parameters.shape))
        loss_function: Function to evaluate the parameters on. It must receive only a
            numpy.ndarray of parameters, and return a real number.
            If your function requires more arguments, consider using the
            'LossFunctionWrapper' class from 'orqviz.loss_function'.
        n_reps: Repetitions to average the output in noisy cases. Defaults to 1.
        verbose: Flag for verbosity of progress. Defaults to False.

    """
    n_points = len(all_points)

    values: List[List[Optional[float]]] = [[None] * n_points] * n_reps
    for rep in range(n_reps):
        for idx, point in enumerate(all_points):
            if idx % 10 == 0 and verbose:
                print("Progress: {:.1f}%".format(round(idx / n_points * 100)))
            values[rep][idx] = loss_function(point)

    return np.array(np.mean(np.asarray(values), axis=0))


def eval_points_on_grid(
    all_parameters: GridOfParameterVectors,
    loss_function: LossFunction,
    n_reps: int = 1,
    verbose: bool = False,
) -> np.ndarray:
    """Function to evaluate loss function on a 2D grid of parameters.

    Args:
        all_parameters:
            Grid of parameters with shape (len_y, len_x, *(parameters.shape))
        loss_function: Function toevaluate the parameters on. It must receive only a
            numpy.ndarray of parameters, and return a real number.
            If your function requires more arguments, consider using the
            'LossFunctionWrapper' class from 'orqviz.loss_function'.
        n_reps: Repetitions to average the output in noisy cases. Defaults to 1.
        verbose: Flag for verbosity of progress. Defaults to False.

    """

    shape = np.shape(all_parameters)
    (size_x, size_y), params_shape = shape[:2], shape[2:]

    vector_of_parameters = all_parameters.reshape((size_x * size_y, *params_shape))

    vector_of_values = eval_points_on_path(
        all_points=vector_of_parameters,
        loss_function=loss_function,
        n_reps=n_reps,
        verbose=verbose,
    )
    return vector_of_values.reshape((size_x, size_y))



from __future__ import annotations

from typing import Callable, NamedTuple, Tuple

import numpy as np
from scipy.interpolate import interp1d

from ..aliases import ArrayOfParameterVectors, LossFunction, ParameterVector, Weights
from ..geometric import _norm_of_arrayofparametervectors
from ..scans import eval_points_on_path


class Chain(NamedTuple):
    """Data structure for Chain in the Nudged Elastic Band (NEB) algorithm.
        This is how we call the NEB with pivot points.

    Args:
        pivots: Array of parameter vectors which form
            a piece-wise linearly connected chain
    """

    pivots: ArrayOfParameterVectors

    def get_weights(self) -> Weights:
        chain_weights = _norm_of_arrayofparametervectors(np.diff(self.pivots, axis=0))
        chain_weights /= np.sum(chain_weights)
        cum_weights = np.cumsum(chain_weights)
        matching_cum_weights = np.insert(cum_weights, 0, 0)
        matching_cum_weights[-1] = 1
        return matching_cum_weights

    def evaluate_on_pivots(self, loss_function: LossFunction) -> np.ndarray:
        return eval_points_on_path(self.pivots, loss_function)

    @property
    def n_pivots(self) -> int:
        return len(self.pivots)

    @property
    def n_params(self) -> int:
        return int(np.prod(self.param_shape))

    @property
    def param_shape(self) -> Tuple[int, ...]:
        return np.atleast_1d(self.pivots[0]).shape


class ChainPath(NamedTuple):
    """Data structure for the piece-wise linear path that is defined by a Chain.
    Args:
        Chain
    """

    primary_chain: Chain

    @property
    def primary_weights(self) -> Weights:
        return self.primary_chain.get_weights()

    def generate_chain(self, n_points: int) -> Chain:
        weight_interpolator = interp1d(
            np.linspace(0, 1, num=len(self.primary_weights)),
            self.primary_weights,
        )
        weights = weight_interpolator(np.linspace(0, 1, num=n_points))
        return self._get_chain_from_weights(weights)

    def generate_uniform_chain(self, n_points: int) -> Chain:
        weights = np.linspace(0, 1, num=n_points)
        return self._get_chain_from_weights(weights)

    def evaluate_points_on_path(
        self, n_points: int, loss_function: LossFunction, weighted: bool = False
    ) -> np.ndarray:
        if weighted:
            chain = self.generate_chain(n_points)
        else:
            chain = self.generate_uniform_chain(n_points)
        return chain.evaluate_on_pivots(loss_function)

    def _get_chain_from_weights(self, weights: Weights) -> Chain:
        distance_between_pivots = np.diff(self.primary_chain.pivots, axis=0)

        chain_diff = np.cumsum(
            _norm_of_arrayofparametervectors(distance_between_pivots)
        )
        chain_diff /= max(chain_diff)
        chain_diff[-1] = 1
        chain_diff = np.insert(chain_diff, 0, 0)

        interpolator = interp1d(
            chain_diff, self.primary_chain.pivots, kind="linear", axis=0
        )

        return Chain(interpolator(weights))



import numpy as np
import pytest
from matplotlib import pyplot as plt

import orqviz

np.random.seed(2)


def loss_function(pars, freq=3.5):
    return np.sum(np.cos(pars * freq))


n_params = 2
params = np.random.uniform(-np.pi, np.pi, size=n_params)
dir1 = np.array([1.0, 0.0])
dir2 = np.array([0.0, 1.0])
end_points = (0, 2 * np.pi)


# Make sure it works for even and odd resolutions
@pytest.mark.parametrize("res", [5, 6])
def test_fourier(res: int):
    fourier_result = orqviz.fourier.scan_2D_fourier(
        params,
        loss_function,
        direction_x=dir1,
        direction_y=dir2,
        n_steps_x=res,
        end_points_x=end_points,
    )

    assert fourier_result.values.shape == (res, res)
    assert fourier_result.end_points_x == fourier_result.end_points_y == end_points


@pytest.mark.parametrize("res", [5, 6])
def test_inverse(res: int):
    scan_2D_result = orqviz.scans.perform_2D_scan(
        params,
        loss_function,
        direction_x=dir1,
        direction_y=dir2,
        n_steps_x=6,  # the inverse of rfft always has an even resolution in the x dir
        n_steps_y=res,
        end_points_x=end_points,
    )
    transformed = orqviz.fourier.perform_2D_fourier_transform(
        scan_2D_result, end_points, end_points
    )
    inversed = orqviz.fourier.inverse_fourier(transformed)

    np.testing.assert_array_almost_equal(scan_2D_result.values, inversed.values)


def test_plots_dont_fail():
    res = 10
    fourier_result = orqviz.fourier.scan_2D_fourier(
        params,
        loss_function,
        direction_x=dir1,
        direction_y=dir2,
        n_steps_x=res,
        end_points_x=end_points,
    )
    orqviz.fourier.plot_2D_fourier_result(fourier_result, max_freq_x=3, max_freq_y=3)
    orqviz.fourier.plot_2D_fourier_result(
        fourier_result, max_freq_x=3, max_freq_y=3, show_full_spectrum=True
    )
    fig, ax = plt.subplots()
    orqviz.fourier.plot_2D_fourier_result(
        fourier_result, max_freq_x=3, max_freq_y=3, fig=fig, ax=ax
    )


def test_inverse_plots_dont_fail():
    res = 5
    fourier_result = orqviz.fourier.scan_2D_fourier(
        params,
        loss_function,
        direction_x=dir1,
        direction_y=dir2,
        n_steps_x=res,
        end_points_x=end_points,
    )
    inverse_result = orqviz.fourier.inverse_fourier(fourier_result)
    orqviz.fourier.plot_inverse_fourier_result(inverse_result)
    fig, ax = plt.subplots()
    orqviz.fourier.plot_inverse_fourier_result(fourier_result, fig, ax)



import os

import matplotlib.pyplot as plt
import numpy as np
import pytest

from orqviz.hessians import (
    get_Hessian,
    get_Hessian_SPSA_approx,
    perform_1D_hessian_eigenvector_scan,
    plot_1D_hessian_eigenvector_scan_result,
)
from orqviz.io import load_viz_object, save_viz_object


def COST_FUNCTION(params):
    return (
        np.sum(np.sin(params)) + np.sum(params**2) + params[0] ** 2 * params[1] ** 2
    )


def ANALYTICAL_HESSIAN(params):
    hessian = np.diag(2 - np.sin(params))
    hessian[0][0] += 2 * params[1] ** 2
    hessian[1][1] += 2 * params[0] ** 2
    hessian[0][1] = 4 * params[0] * params[1]
    hessian[1][0] = 4 * params[0] * params[1]
    return hessian


@pytest.mark.parametrize(
    "params",
    [np.random.rand(8)],
)
def test_get_hessian_io(params):
    hessian = get_Hessian(params, COST_FUNCTION, gradient_function=None, eps=1e-3)

    assert hasattr(hessian, "eigenvectors")
    assert hasattr(hessian, "eigenvalues")
    assert (
        len(hessian.eigenvalues)
        == len(hessian.eigenvectors)
        == len(hessian.eigenvectors.T)
        == len(params)
    )

    save_viz_object(hessian, "test")
    loaded_hessian = load_viz_object("test")
    os.remove("test")
    np.testing.assert_array_almost_equal(
        loaded_hessian.eigenvalues, hessian.eigenvalues
    )
    np.testing.assert_array_almost_equal(
        loaded_hessian.eigenvectors, hessian.eigenvectors
    )


@pytest.mark.parametrize(
    "params",
    [np.random.rand(8)],
)
def test_get_hessian_SPSA_approx_io(params):
    hessian = get_Hessian_SPSA_approx(
        params, COST_FUNCTION, gradient_function=None, eps=1e-3, n_reps=20
    )

    assert hasattr(hessian, "eigenvectors")
    assert hasattr(hessian, "eigenvalues")
    assert (
        len(hessian.eigenvalues)
        == len(hessian.eigenvectors)
        == len(hessian.eigenvectors.T)
        == len(params)
    )
    save_viz_object(hessian, "test")
    loaded_hessian = load_viz_object("test")
    os.remove("test")
    np.testing.assert_array_almost_equal(
        loaded_hessian.eigenvalues, hessian.eigenvalues
    )
    np.testing.assert_array_almost_equal(
        loaded_hessian.eigenvectors, hessian.eigenvectors
    )


HESSIAN_TEST_PARAMS = [
    np.zeros(4),
    np.ones(5),
    np.arange(4, dtype=float),
    np.pi / 4 * np.arange(8),
]


@pytest.mark.parametrize("params", HESSIAN_TEST_PARAMS)
def test_get_hessian_gives_correct_values(params):
    eps = 1e-5
    target_matrix = ANALYTICAL_HESSIAN(params)
    target_eigenvalues = np.sort(np.linalg.eig(target_matrix)[0])

    hessian = get_Hessian(params, COST_FUNCTION, gradient_function=None, eps=eps)
    precision = int(np.abs(np.log10(eps))) - 1

    np.testing.assert_array_almost_equal(
        hessian.hessian_matrix, target_matrix, precision
    )
    np.testing.assert_array_almost_equal(
        hessian.eigenvalues, target_eigenvalues, precision
    )


@pytest.mark.parametrize("params", HESSIAN_TEST_PARAMS)
def test_get_hessian_spsa_gives_correct_values(params):
    eps = 1e-5
    target_matrix = ANALYTICAL_HESSIAN(params)
    target_eigenvalues = np.sort(np.linalg.eig(target_matrix)[0])

    hessian = get_Hessian_SPSA_approx(
        params, COST_FUNCTION, gradient_function=None, eps=eps, n_reps=20000
    )
    precision = 1
    np.testing.assert_array_almost_equal(
        hessian.hessian_matrix, target_matrix, precision
    )
    np.testing.assert_array_almost_equal(
        hessian.eigenvalues, target_eigenvalues, precision
    )



import pickle
from typing import Union

from .elastic_band.data_structures import Chain
from .hessians.data_structures import HessianEigenobject
from .scans.data_structures import Scan1DResult, Scan2DResult

OrqVizObject = Union[Scan1DResult, Scan2DResult, HessianEigenobject, Chain]


def save_viz_object(viz_object: OrqVizObject, filename: str) -> None:
    """Save datatype to a file with pickle"""

    with open(filename, "wb") as f:
        pickle.dump(viz_object, f)


def load_viz_object(filename: str) -> OrqVizObject:
    """Load datatype from a file with pickle"""

    with open(filename, "rb") as f:
        loaded_object = pickle.load(f)

    return loaded_object



import numpy as np
import pytest

from orqviz.gradients import calculate_full_gradient


def SUM_OF_SINES(params):
    return np.sum(np.sin(params))


def dSUM_OF_SINES_dPARAMS(params):
    return np.cos(params)


@pytest.mark.parametrize(
    "params",
    [
        np.array([1, 2, -3, 4, -1]),
        np.array([0.123, -0.987, 1, 6, 18, -23.8]),
        np.array([-np.pi, np.pi / 7, np.pi * 2]),
    ],
)
def test_numerical_gradient(params):

    gradient_values = calculate_full_gradient(
        params, SUM_OF_SINES, stochastic=False, eps=1e-6
    )
    assert gradient_values.shape == np.shape(params)
    assert np.allclose(gradient_values, dSUM_OF_SINES_dPARAMS(params))


@pytest.mark.parametrize(
    "params",
    [
        np.array([1, 2, -3, 4, -1]),
        np.array([0.123, -0.987, 1, 6, 18, -23.8]),
        np.array([-np.pi, np.pi / 7, np.pi * 2]),
    ],
)
def test_stochastic_gradient(params):

    gradient_values = np.mean(
        [
            calculate_full_gradient(
                params,
                SUM_OF_SINES,
                stochastic=True,
                eps=1e-6,
            )
            for _ in range(10000)
        ],
        axis=0,
    )
    assert gradient_values.shape == np.shape(params)
    assert np.allclose(gradient_values, dSUM_OF_SINES_dPARAMS(params), atol=0.1)



import os

import matplotlib.pyplot as plt
import numpy as np

from orqviz.hessians import (
    get_Hessian,
    perform_1D_hessian_eigenvector_scan,
    plot_1D_hessian_eigenvector_scan_result,
)
from orqviz.io import load_viz_object, save_viz_object


def COST_FUNCTION(params):
    return np.sum(np.sin(params)) + np.sum(params**2) + 3 * params[1] - 10 * params[2]


def test_plot_1D_hessian_eigenvector_scan_result():
    params = np.random.rand(8)

    hessian = get_Hessian(params, COST_FUNCTION, gradient_function=None, eps=1e-3)

    list_of_scans = perform_1D_hessian_eigenvector_scan(hessian, COST_FUNCTION)
    plot_1D_hessian_eigenvector_scan_result(list_of_scans, hessian.eigenvalues)
    fig, ax = plt.subplots()
    plot_1D_hessian_eigenvector_scan_result(list_of_scans, hessian.eigenvalues, ax=ax)



from typing import Callable, List, Optional, Tuple

import numpy as np

from ..aliases import GradientFunction, LossFunction, ParameterVector
from ..gradients import numerical_gradient
from ..scans.data_structures import Scan1DResult
from ..scans.scans_1D import perform_1D_scan
from .data_structures import HessianEigenobject


def perform_1D_hessian_eigenvector_scan(
    hessian_object: HessianEigenobject,
    loss_function: LossFunction,
    n_points: int = 31,
    endpoints: Tuple[float, float] = (-np.pi, np.pi),
) -> List[Scan1DResult]:
    """Function to create 1D scans along all eigenvecor directions
        of a Hessian and return a list of scans.

    Args:
        hessian_object: HessianEigenobject Datatype containing a Hessian matrix.
        loss_function: Function to perform the scan on. It must receive only a
            numpy.ndarray of parameters, and return a real number.
            If your function requires more arguments, consider using the
            'LossFunctionWrapper' class from 'orqviz.loss_function'.
        n_points: Number of points to evaluate the loss along each direction.
            Defaults to 31.
        endpoints: End points for scan along each direction.
            Defaults to (-np.pi, np.pi).
    """
    list_of_scans = []
    for direction in hessian_object.eigenvectors:
        scan = perform_1D_scan(
            hessian_object.params, loss_function, direction, n_points, endpoints
        )
        list_of_scans.append(scan)

    return list_of_scans


def get_Hessian(
    params: ParameterVector,
    loss_function: LossFunction,
    gradient_function: Optional[GradientFunction] = None,
    n_reps: int = 1,
    eps: float = 0.1,
) -> HessianEigenobject:
    """Function to calculate the Hessian matrix of a loss function
        (matrix of second-order partial derivatives) at a specified parameter vector.
    The calculation is performed by combining an optional gradient function
        and finite difference gradients.
    Calculation time scales quadratically with the number of parameters.

    Args:
        params: Parameter vector at which the Hessian matrix is computed.
        loss_function: Function to calculate the Hessian of. It must receive only a
            numpy.ndarray of parameters, and return a real number.
            If your function requires more arguments, consider using the
            'LossFunctionWrapper' class from 'orqviz.loss_function'.
        gradient_function: Gradient function which can be used to calculate
            the partial derivative of the loss function for individial parameters.
            It can be used to avoid some numerical gradients and improve
            accuracy of the Hessian. Defaults to None.
        n_reps: Number of repetitions over which Hessian is averaged.
            Useful if noise is present. Defaults to 1.
        eps: Finite difference stencil used for numerical gradient in calculation.
            It is always used, even if gradient function is provided. Defaults to 0.1.
    """

    flat_params = params.flatten()

    n_params = len(flat_params)
    hessian_shape = (n_params, n_params)
    Hessian_Matr = np.zeros(shape=hessian_shape)

    if gradient_function is None:

        def _gradient_function(x: ParameterVector, d: ParameterVector) -> float:
            return numerical_gradient(x, d, loss_function, eps)

        gradient_function = _gradient_function

    for _ in range(n_reps):
        for j in range(n_params):
            dir1 = np.zeros_like(flat_params)
            dir1[j] = 1
            dir1 = dir1.reshape(params.shape)
            dir1_gradient = gradient_function(params, dir1)
            for k in range(0, j + 1):
                dir2 = np.zeros_like(flat_params)
                dir2[k] = 1
                dir2 = dir2.reshape(params.shape)
                dir2_gradient = gradient_function(params + dir2 * eps, dir1)

                op = np.outer(dir1.flatten(), dir2.flatten())
                outer_prod_matrix = op + op.T
                # half diagonal entries
                outer_prod_matrix[(np.arange(n_params), np.arange(n_params))] /= 2

                hessian_result = (dir2_gradient - dir1_gradient) / eps
                Hessian_Matr += hessian_result * outer_prod_matrix

    return HessianEigenobject(params, Hessian_Matr / n_reps)


def get_Hessian_SPSA_approx(
    params: ParameterVector,
    loss_function: LossFunction,
    gradient_function: Optional[GradientFunction] = None,
    n_reps: int = 20,
    eps: float = 0.1,
) -> HessianEigenobject:
    """Function to calculate an SPSA approximation of the the Hessian matrix of
        a loss function (matrix of second-order partial derivatives)
        at a specified parameter vector.
    The Hessian is approximated by estimating the second partial derivative
        in random stochastic directions.

    Args:
        params: Parameter vector at which the Hessian matrix is computed.
        loss_function: Function to calculate the Hessian of. It must receive only a
            numpy.ndarray of parameters, and return a real number.
            If your function requires more arguments, consider using the
            'LossFunctionWrapper' class from 'orqviz.loss_function'.
        gradient_function: Gradient function which can be used to calculate
            the derivative of the loss function in random stochastic directions.
            It can be used to avoid some numerical gradients and improve accuracy
            of the Hessian. Defaults to None.
        n_reps: Number of random stochastic gradients used to approximate the Hessian.
            Defaults to 20.
        eps: Finite difference stencil used for numerical gradient in calculation.
            It is always used, even if gradient function is provided. Defaults to 0.1.
    """

    flat_params = params.flatten()

    n_params = len(flat_params)
    hessian_shape = (n_params, n_params)
    Hessian_Matr = np.zeros(shape=hessian_shape)

    if gradient_function is None:

        def _gradient_function(x: ParameterVector, d: ParameterVector) -> float:
            return numerical_gradient(x, d, loss_function, eps)

        gradient_function = _gradient_function

    for _ in range(n_reps):
        dir1 = np.random.choice([-1.0, 1.0], size=params.shape)
        dir1_gradient = gradient_function(params, dir1)

        dir2 = np.random.choice([-1.0, 1.0], size=params.shape)
        dir2_gradient = gradient_function(params + dir2 * eps, dir1)

        op = np.outer(dir1.flatten(), dir2.flatten())
        outer_prod_matrix = (op + op.T) / 2

        hessian_result = (dir2_gradient - dir1_gradient) / eps
        Hessian_Matr += hessian_result * outer_prod_matrix

    return HessianEigenobject(params, Hessian_Matr / n_reps)



import matplotlib.pyplot as plt
import numpy as np

from orqviz.elastic_band import Chain, plot_all_chains_losses, run_AutoNEB


def SUM_OF_SINES(params):
    return np.sum(np.sin(params))


def test_plot_all_chains_losses():
    pivots = np.array([[t, 1 - t] for t in np.linspace(0, 1, 10)])
    init_chain = Chain(pivots=pivots)

    n_cycles = 4
    max_new_pivots = 2
    all_chains = run_AutoNEB(
        init_chain,
        loss_function=SUM_OF_SINES,
        full_gradient_function=None,
        n_cycles=n_cycles,
        n_iters_per_cycle=10,
        max_new_pivots=max_new_pivots,
    )

    plot_all_chains_losses(all_chains, SUM_OF_SINES)
    fig, ax = plt.subplots()
    plot_all_chains_losses(all_chains, SUM_OF_SINES, ax=ax)



import matplotlib.pyplot as plt
import numpy as np
import pytest

from orqviz.pca import get_pca, perform_2D_pca_scan
from orqviz.scans.data_structures import Scan2DResult


def COST_FUNCTION(params):
    return np.sum(np.sin(params)) + np.sum(params**2)


@pytest.mark.parametrize(
    "all_points,components_ids",
    [[np.random.rand(98, 9), (0, 1)], [np.random.rand(27, 10, 4), (1, 5)]],
)
def test_pca(all_points, components_ids):
    pca = get_pca(all_points, components_ids=components_ids)

    n_steps_x = 100
    n_steps_y = 20

    scan_results = perform_2D_pca_scan(
        pca_object=pca,
        loss_function=COST_FUNCTION,
        n_steps_x=n_steps_x,
        n_steps_y=n_steps_y,
    )
    assert scan_results.params_grid.shape == (n_steps_y, n_steps_x, 2)
    np.testing.assert_array_equal(scan_results.direction_x, np.array([1, 0]))
    np.testing.assert_array_equal(scan_results.direction_y, np.array([0, 1]))
    assert pca.get_transformed_points().shape == (
        len(all_points),
        max(pca.components_ids) + 1,
    )
    assert pca.get_transformed_points(all_points[:24]).shape == (
        24,
        max(pca.components_ids) + 1,
    )

    new_components_ids = (2, 3)
    pca.set_component_ids(new_components_ids)

    scan_results = perform_2D_pca_scan(
        pca_object=pca,
        loss_function=COST_FUNCTION,
        n_steps_x=n_steps_x,
        n_steps_y=n_steps_y,
    )
    assert scan_results.params_grid.shape == (n_steps_y, n_steps_x, 2)
    np.testing.assert_array_equal(scan_results.direction_x, np.array([1, 0]))
    np.testing.assert_array_equal(scan_results.direction_y, np.array([0, 1]))
    assert pca.get_transformed_points().shape == (
        len(all_points),
        max(pca.components_ids) + 1,
    )
    assert pca.get_transformed_points(all_points[:24]).shape == (
        24,
        max(pca.components_ids) + 1,
    )



from typing import Optional

import matplotlib
import numpy as np

from ..plot_utils import _check_and_create_3D_ax, _check_and_create_fig_ax
from .data_structures import Scan1DResult, Scan2DResult


def plot_1D_scan_result(
    scan1d_result: Scan1DResult,
    ax: Optional[matplotlib.axes.Axes] = None,
    in_units_of_direction: bool = False,
    **plot_kwargs,
) -> None:
    """Function to plot a Scan1DResult.

    Args:
        scan1d_result: Scan result to be plotted.
        ax: Matplotlib axis to perform plot on.
            If None, a new axis is created from the current figure. Defaults to None.
        in_units_of_direction: Flag to indicate axis ticks are in units
            of the direction. If False, ticks are in units of Euclidean distance.
            Defaults to False.
        plot_kwargs: kwargs for plotting with matplotlib.pyplot.plot (plt.plot)
    """
    _, ax = _check_and_create_fig_ax(ax=ax)

    default_plot_kwargs = {"linewidth": 3}
    plot_kwargs = {**default_plot_kwargs, **plot_kwargs}

    ax.plot(
        scan1d_result._get_coordinates_on_direction(
            in_units_of_direction=in_units_of_direction
        ),
        scan1d_result.values,
        **plot_kwargs,
    )
    ax.set_xlabel("Scan Direction")
    ax.set_ylabel("Loss Value")


def plot_1D_interpolation_result(
    interpolation_result: Scan1DResult,
    ax: Optional[matplotlib.axes.Axes] = None,
    **plot_kwargs,
) -> None:
    """Function to plot a Scan1DResult of a 1D interpolation.

    Args:
        interpolation_result: Interpolation scan result to be plotted.
        ax: Matplotlib axis to perform plot on. If None, a new axis is created
        from the current figure. Defaults to None.
        plot_kwargs: kwargs for plotting with matplotlib.pyplot.plot (plt.plot)
    """
    _, ax = _check_and_create_fig_ax(ax=ax)

    default_plot_kwargs = {"color": "royalblue"}
    plot_kwargs = {**default_plot_kwargs, **plot_kwargs}

    plot_1D_scan_result(
        scan1d_result=interpolation_result,
        ax=ax,
        in_units_of_direction=True,
        **plot_kwargs,
    )
    ax.axvline(x=0.0, color="red")
    ax.axvline(x=1.0, color="red")
    ax.set_xlabel("Interpolation Direction")
    ax.set_ylabel("Loss Value")


def plot_2D_scan_result(
    scan2d_result: Scan2DResult,
    fig: Optional[matplotlib.figure.Figure] = None,
    ax: Optional[matplotlib.axes.Axes] = None,
    in_units_of_direction: bool = False,
    **plot_kwargs,
) -> None:
    """Function to plot a Scan2DResult.

    Args:
        scan2d_result: Scan result to be plotted.
        fig: Matplotlib figure to perfom a plot on.
            If None, a new figure and axis are created. Defaults to None.
        ax: Matplotlib axis to perform plot on. If None, a new axis is created
            from the current figure. Defaults to None.
        in_units_of_direction: Flag to indicate axis ticks are in units
            of the direction. If False, ticks are in units of Euclidean distance.
            Defaults to False.
        plot_kwargs: kwargs for plotting with matplotlib.pyplot.pcolormesh
            (plt.pcolormesh)
    """
    fig, ax = _check_and_create_fig_ax(fig=fig, ax=ax)
    x, y = scan2d_result._get_coordinates_on_directions(
        in_units_of_direction=in_units_of_direction
    )
    XX, YY = np.meshgrid(x, y)

    default_plot_kwargs = {"shading": "auto"}
    plot_kwargs = {**default_plot_kwargs, **plot_kwargs}

    mesh_plot = ax.pcolormesh(
        XX, YY, scan2d_result.values, **plot_kwargs, rasterized=True
    )
    fig.colorbar(mesh_plot, ax=ax)
    ax.set_xlabel("Scan Direction x")
    ax.set_ylabel("Scan Direction y")


def plot_2D_interpolation_result(
    scan2D_result: Scan2DResult,
    fig: Optional[matplotlib.figure.Figure] = None,
    ax: Optional[matplotlib.axes.Axes] = None,
    **plot_kwargs,
) -> None:
    """Function to plot a Scan2DResult from 2D interpolation.

    Args:
        scan2d_result: Scan result to be plotted.
        fig: Matplotlib figure to perfom a plot on. If None, a new figure
            and axis are created from the current figure. Defaults to None.
        ax: Matplotlib axis to perform plot on. If None, a new axis is created.
            Defaults to None.
        plot_kwargs: kwargs for plotting with matplotlib.pyplot.pcolormesh
            (plt.pcolormesh)
    """
    fig, ax = _check_and_create_fig_ax(fig=fig, ax=ax)

    plot_2D_scan_result(
        scan2d_result=scan2D_result,
        fig=fig,
        ax=ax,
        in_units_of_direction=True,
        **plot_kwargs,
    )
    ax.scatter(0, 0, color="red", linewidth=2)
    ax.scatter(1, 0, color="red", linewidth=2)
    ax.set_xlabel("Interpolation Direction")
    ax.set_ylabel("Scan Direction y")


def plot_2D_scan_result_as_3D(
    scan2D_result: Scan2DResult,
    ax: Optional[matplotlib.axes.Axes] = None,
    in_units_of_direction: bool = False,
    **plot_kwargs,
) -> None:
    """Function to create a 3D plot from a Scan2DResult.

    Args:
        scan2d_result: Scan result to be plotted.
        ax: Matplotlib axis to perform plot on. Must have projection='3d'
            set on creation. If None, a new axis is created. Defaults to None.
        in_units_of_direction: Flag to indicate axis ticks are in units
            of the direction. If False, ticks are in units of Euclidean distance.
            Defaults to False.
        plot_kwargs: kwargs for plotting with matplotlib.pyplot.plot_surface
            (plt.plot_surface)
    """
    ax3D = _check_and_create_3D_ax(ax=ax)

    assert ax3D is not None

    x, y = scan2D_result._get_coordinates_on_directions(
        in_units_of_direction=in_units_of_direction
    )
    XX, YY = np.meshgrid(x, y)

    plot_kwargs_defaults = {"cmap": "viridis", "alpha": 0.8}
    plot_kwargs = {**plot_kwargs_defaults, **plot_kwargs}

    ax3D.plot_surface(XX, YY, scan2D_result.values, **plot_kwargs)

    ax3D.view_init(elev=35, azim=-70)

    ax3D.set_xlabel("Scan Direction x")
    ax3D.set_ylabel("Scan Direction y")
    ax3D.set_zlabel("Loss Value")



from typing import Callable, List, Optional, Tuple, Union

import numpy as np

from ..aliases import LossFunction, ParameterVector
from ..scans import Scan2DResult, perform_2D_scan
from .data_structures import PCAobject


def perform_2D_pca_scan(
    pca_object: PCAobject,
    loss_function: LossFunction,
    n_steps_x: int = 20,
    n_steps_y: Optional[int] = None,
    offset: Union[Tuple[float, float], float] = (-1.0, 1.0),
    verbose: bool = False,
) -> Scan2DResult:
    """Function to perform a 2D scan on a loss function landscape
        according in directions of PCA components.

    Args:
        all_points: Points on which PCA was fitted and around which the scan
            is performed.
        loss_function: Function to perform the scan on. It must receive only a
            numpy.ndarray of parameters, and return a real number.
            If your function requires more arguments, consider using the
            'LossFunctionWrapper' class from 'orqviz.loss_function'.
        pca: PCA object that was fitted on all_points.
            Its components are used to decide scan directions.
        components_ids: Which components of the PCA object are used as scan directions.
            Defaults to (0, 1).
        n_steps_x: Number of grid points in x-direction to perform the scan.
            Defaults to 20.
        n_steps_y: Number of grid points in y-direction to perform the scan.
            If set to None, it is set to n_steps_x. Defaults to None.
        offset: Offset in x-y directions added to the scan range on top of
            the range that is necessary to display all_points. Defaults to (-1.0, 1.0).
        verbose: Flag for printing progress. Defaults to False.
    """

    if not isinstance(offset, tuple):
        offset = (-np.abs(offset), np.abs(offset))

    if n_steps_y is None:
        n_steps_y = n_steps_x

    end_points_x, end_points_y = pca_object._get_endpoints_from_pca(offset)

    def pca_loss_function(xy_params):
        pca_params = np.zeros(pca_object.pca.n_components)
        pca_params[pca_object.components_ids[0]] = xy_params[0]
        pca_params[pca_object.components_ids[1]] = xy_params[1]
        probe_parameters = pca_object.get_inverse_transformed_point(pca_params)
        return loss_function(probe_parameters)

    return perform_2D_scan(
        origin=np.array([0, 0]),
        loss_function=pca_loss_function,
        direction_x=np.array([1, 0]),
        direction_y=np.array([0, 1]),
        n_steps_x=n_steps_x,
        n_steps_y=n_steps_y,
        end_points_x=end_points_x,
        end_points_y=end_points_y,
        verbose=verbose,
    )



from .data_structures import (
    Scan1DResult,
    Scan2DResult,
    clone_Scan1DResult_with_different_values,
    clone_Scan2DResult_with_different_values,
)
from .evals import eval_points_on_grid, eval_points_on_path
from .plots import (
    plot_1D_interpolation_result,
    plot_1D_scan_result,
    plot_2D_interpolation_result,
    plot_2D_scan_result,
    plot_2D_scan_result_as_3D,
)
from .scans_1D import perform_1D_interpolation, perform_1D_scan
from .scans_2D import perform_2D_interpolation, perform_2D_scan



from .data_structures import PCAobject, get_pca
from .plots import (
    plot_line_through_points_on_pca,
    plot_optimization_trajectory_on_pca,
    plot_pca_landscape,
    plot_scatter_points_on_pca,
)
from .scans import perform_2D_pca_scan



from dataclasses import dataclass
from typing import List, Optional, Tuple

import numpy as np
from sklearn.decomposition import PCA

from ..aliases import ArrayOfParameterVectors


@dataclass(init=True)
class PCAobject:
    """PCA datatype to combine PCA object with the corresponding higher-dimensional
    points and the components of interest"""

    def __init__(
        self,
        all_points: ArrayOfParameterVectors,
        components_ids: Tuple[int, int] = (0, 1),
    ):
        self.all_points = np.array(all_points)
        self._params_shape = np.shape(self.all_points)[1:]
        self.components_ids = components_ids
        self.fit_pca(max(self.components_ids) + 1)

    def set_component_ids(self, new_component_ids: Tuple[int, int]) -> None:
        new_n_components = max(new_component_ids) + 1
        self.fit_pca(new_n_components)
        self.components_ids = new_component_ids

    def fit_pca(self, n_components: int) -> None:
        self.pca = PCA(n_components=n_components)
        params = self.all_points.reshape(-1, int(np.prod(self._params_shape)))
        self.pca.fit(params)

    def get_transformed_points(
        self, points: Optional[ArrayOfParameterVectors] = None
    ) -> np.ndarray:
        high_dim_points = self.all_points if points is None else points
        return self.pca.transform(
            high_dim_points.reshape(-1, int(np.prod(self._params_shape)))
        )

    def get_inverse_transformed_point(
        self, pca_params: np.ndarray
    ) -> ArrayOfParameterVectors:
        return self.pca.inverse_transform(pca_params).reshape(*self._params_shape)

    @property
    def components(self) -> np.ndarray:
        """Returns a numpy array of component vectors in the shape of the parameters"""
        return self.pca.components_.reshape(-1, *self._params_shape)

    @property
    def mean(self) -> np.ndarray:
        """Returns the fitted mean of the high-dimensional parameter vectors"""
        return self.pca.mean_.reshape(*self._params_shape)

    def _get_endpoints_from_pca(
        self,
        offset: Tuple[float, float] = (-1.0, 1.0),
    ) -> Tuple[Tuple[float, float], Tuple[float, float]]:
        """Helper function to get the scan coordinate ranges of PCA-transformed
            coordinates in the specified components with a provided offset.

        Args:
            components_ids: Which components of the PCA object are used as
                scan directions. Defaults to (0, 1).
            offset: Offset in x-y directions added to the scan range on top
                of the range that is necessary to display all_points.
                Defaults to (-1.0, 1.0).

        """
        pca_transformed_points = self.get_transformed_points()

        end_points_x = (
            float(min(pca_transformed_points[:, self.components_ids[0]]) + offset[0]),
            float(max(pca_transformed_points[:, self.components_ids[0]]) + offset[1]),
        )
        end_points_y = (
            float(min(pca_transformed_points[:, self.components_ids[1]]) + offset[0]),
            float(max(pca_transformed_points[:, self.components_ids[1]]) + offset[1]),
        )
        return end_points_x, end_points_y


def get_pca(
    all_points: ArrayOfParameterVectors,
    components_ids: Tuple[int, int] = (0, 1),
) -> PCAobject:
    """Fits and returns a sklearn PCA instance to the provided parameters.

    Args:
        all_points: List/array of parameter vectors to perform PCA on
        components_ids: Which components are of interest. PCA fits to
            max(components_ids)+1 components. Defaults to (0, 1).
    """
    return PCAobject(all_points, components_ids)



import math
import warnings
from typing import NamedTuple, Optional, Tuple

import matplotlib
import numpy as np
from matplotlib.ticker import MaxNLocator

from orqviz.scans.data_structures import Scan2DResult

from .aliases import DirectionVector, LossFunction, ParameterVector
from .plot_utils import _check_and_create_fig_ax
from .scans.scans_2D import perform_2D_scan


class FourierResult(NamedTuple):
    """Datatype for 2D Fourier scans to combine the scan result and scan instruction."""

    values: np.ndarray
    end_points_x: Tuple[float, float]
    end_points_y: Tuple[float, float]


def scan_2D_fourier(
    origin: ParameterVector,
    loss_function: LossFunction,
    direction_x: Optional[DirectionVector] = None,
    direction_y: Optional[DirectionVector] = None,
    n_steps_x: int = 20,
    n_steps_y: Optional[int] = None,
    end_points_x: Tuple[float, float] = (0, 2 * np.pi),
    end_points_y: Optional[Tuple[float, float]] = None,
) -> FourierResult:
    """Performs a discrete real fourier transform on the 2D scan of a loss function.

    Args:
        origin: Origin point of the 2D scan.
        loss_function: Function to perform the scan on. It must receive only a
            numpy.ndarray of parameters, and return a real number.
            If your function requires more arguments, consider using the
            'LossFunctionWrapper' class from 'orqviz.loss_function'.
        direction_x: x-direction vector for scan. Has same shape as origin.
            If None, a random unit vector is sampled. Defaults to None.
        direction_y: y-direction vector for scan. Has same shape as origin.
            If None, a random unit vector is sampled. Defaults to None.
        n_steps_x: Number of points evaluated along the x-direction. Defaults to 20.
        n_steps_y: Number of points evaluated along the y-direction.
            If None, set value to n_steps_x. Defaults to None.
        end_points_x: Range of scan along the x-direction in units of direction_x.
            Defaults to (0, 2pi).
        end_points_y: Range of scan along the y-direction in units of direction_y.
            Defaults to (0, 2pi).

    Returns:
        FourierResult with the following attributes:
            values: Output from np.fft.rfft2 with the format of numpy's output, which is
                the coefficient for 0, then the coefficients of positive frequencies in
                increasing order by frequency, then the coefficients of negative
                frequences in increasing order.
            end_points_x: End points of the scan along the x-direction.
            end_points_y: End points of the scan along the y-direction.
    """
    if end_points_y is None:
        end_points_y = end_points_x
    scan2D_result = perform_2D_scan(
        origin,
        loss_function,
        direction_x=direction_x,
        direction_y=direction_y,
        n_steps_x=n_steps_x,
        n_steps_y=n_steps_y,
        end_points_x=end_points_x,
        end_points_y=end_points_y,
    )
    return perform_2D_fourier_transform(scan2D_result, end_points_x, end_points_y)


def perform_2D_fourier_transform(
    scan2D_result: Scan2DResult,
    end_points_x: Tuple[float, float],
    end_points_y: Tuple[float, float],
):
    """Performs a discrete real fourier transform on an already completed scan of a
    loss function.

    Args:
        scan2D_result: Result of a 2D scan. Output of orqviz.scans.perform_2D_scan.
        end_points_x: Range used for the scan along the x-direction.
        end_points_y: Range used for the scan along they-direction.
    """
    fourier_result = np.fft.fftshift(np.fft.fftn(scan2D_result.values, norm="forward"))
    return FourierResult(fourier_result, end_points_x, end_points_y)


def plot_2D_fourier_result(
    fourier_result: FourierResult,
    max_freq_x: int,
    max_freq_y: int,
    show_full_spectrum: bool = False,
    remove_constant_term: bool = True,
    fig: Optional[matplotlib.figure.Figure] = None,
    ax: Optional[matplotlib.axes.Axes] = None,
    **plot_kwargs,
):
    """Plots a 2D fourier result.
    Note: Fourier coefficients are complex numbers. However, due to the inability to
    plot complex numbers on real axes, the magnitudes of the coefficients are plotted.
    This means that phase has no influence on the visual output of the plot.

    Note: If the plot appears distorted, most often the reason for that is the choice
    of max_freq_x or max_freq_y.

    Args:
        fourier_result: Fourier result to be plotted.
        max_freq_x: Maximum frequency to be plotted along the x-direction.
        max_freq_y: Maximum frequency to be plotted along the y-direction.
        show_full_spectrum: If true plots the whole spectrum (including the
            redundant part). Defaults to False.
        remove_constant_term: If true sets the value of (0, 0) frequency (the constant
            term) to 0. Defaults to True.
        fig: Matplotlib figure to perfom a plot on. If None, a new figure
            and axis are created from the current figure. Defaults to None.
        ax: Matplotlib axis to perform plot on. If None, a new axis is created.
            Defaults to None.
        plot_kwargs: kwargs for plotting with matplotlib.pyplot.pcolormesh
            (plt.pcolormesh)
    """
    raw_values = fourier_result.values
    size = raw_values.shape[0]

    half = int(size / 2)
    if show_full_spectrum:
        min_x = int(half - max_freq_x)
    else:
        min_x = half
    max_x = int(half + max_freq_x)
    min_y = int(half - max_freq_y) - 1
    max_y = int(half + max_freq_y) + 1
    result = fourier_result
    plottable_result = np.abs(result.values)

    if remove_constant_term:
        plottable_result[half][half] = 0
    truncated_result = plottable_result[min_y:max_y, min_x:max_x]

    n_x = truncated_result.shape[1]
    n_y = truncated_result.shape[0]

    if show_full_spectrum:
        x_axis = np.arange(-n_x / 2, n_x / 2)
    else:
        x_axis = np.arange(0, n_x)
    y_axis = np.arange(-n_y / 2, n_y / 2)

    XX, YY = np.meshgrid(x_axis, y_axis)

    default_plot_kwargs = {"shading": "auto"}
    plot_kwargs = {**default_plot_kwargs, **plot_kwargs}
    fig, ax = _check_and_create_fig_ax(fig=fig, ax=ax)
    mesh_plot = ax.pcolormesh(XX, YY, truncated_result, **plot_kwargs, rasterized=True)

    fig.colorbar(mesh_plot, ax=ax)
    ax.set_xlabel("Frequencies x")
    ax.set_ylabel("Frequencies y")
    ax.set_ylim(-max_freq_y, max_freq_y)
    ax.xaxis.set_major_locator(MaxNLocator(integer=True))
    ax.yaxis.set_major_locator(MaxNLocator(integer=True))


def inverse_fourier(result: FourierResult) -> FourierResult:
    """Inverts a Fourier result.
    Args:
        result: Fourier result to be inverted.
    Returns:
        Inverted result.
    """
    return FourierResult(
        np.fft.ifftn(np.fft.ifftshift(result.values), norm="forward"),
        result.end_points_x,
        result.end_points_y,
    )


def plot_inverse_fourier_result(
    result: FourierResult,
    fig=None,
    ax=None,
    **plot_kwargs,
):
    """Plots inverse fourier result according to end points.
    Args:
        result: output of `orqviz.fourier.inverse_fourier`
    """
    Nx = result.values.shape[1]
    Ny = result.values.shape[0]
    x_axis = (
        np.arange(Nx) / Nx * (result.end_points_x[1] - result.end_points_x[0])
        + result.end_points_x[0]
    )
    y_axis = (
        np.arange(Ny) / Ny * (result.end_points_y[1] - result.end_points_y[0])
        + result.end_points_y[0]
    )
    XX, YY = np.meshgrid(x_axis, y_axis)

    default_plot_kwargs = {"shading": "auto"}
    plot_kwargs = {**default_plot_kwargs, **plot_kwargs}

    fig, ax = _check_and_create_fig_ax(fig=fig, ax=ax)
    # if it was rfft then only the real components matter.
    plottable_result = np.real(result.values)
    mesh_plot = ax.pcolormesh(XX, YY, plottable_result, **plot_kwargs, rasterized=True)
    fig.colorbar(mesh_plot, ax=ax)
    ax.set_xlabel("Scan Direction x")
    ax.set_ylabel("Scan Direction y")



from typing import Optional

import matplotlib
import numpy as np

from ..aliases import ArrayOfParameterVectors
from ..plot_utils import _check_and_create_fig_ax
from ..plots import (
    plot_line_through_points,
    plot_optimization_trajectory,
    plot_scatter_points,
)
from ..scans import plot_2D_scan_result
from ..scans.data_structures import Scan2DResult
from .data_structures import PCAobject


def plot_optimization_trajectory_on_pca(
    optimization_trajectory: ArrayOfParameterVectors,
    pca_object: PCAobject,
    ax: Optional[matplotlib.axes.Axes] = None,
    **plot_kwargs,
):
    """Wrapper function around plot_optimization_trajectory to simplify
        plotting a on PCA scan.

    Args:
        optimization_trajectory: Parameter trajectory to be projected on a PCA scan.
        pca_object: PCAobject the contains a fitted PCA object, corresponding points
            and the components of interest.
        ax: Matplotlib axis to perform the plot on. If None, an axis is created
            from the current figure. Defaults to None.
        plot_kwargs: kwargs for plotting with matplotlib.pyplot.plot (plt.plot)
    """

    _, ax = _check_and_create_fig_ax(ax=ax)

    plot_optimization_trajectory(
        optimization_trajectory=optimization_trajectory,
        direction_x=pca_object.components[pca_object.components_ids[0]],
        direction_y=pca_object.components[pca_object.components_ids[1]],
        shift=pca_object.mean,
        ax=ax,
        **plot_kwargs,
    )


def plot_scatter_points_on_pca(
    scatter_points: ArrayOfParameterVectors,
    pca_object: PCAobject,
    ax: Optional[matplotlib.axes.Axes] = None,
    **plot_kwargs,
):
    """Wrapper function around plot_scatter_points to simplify plotting a on PCA scan.

    Args:
        scatter_points: Parameter points to be projected on a PCA scan.
        pca_object: PCAobject the contains a fitted PCA object, corresponding points
            and the components of interest.
        ax: Matplotlib axis to perform the plot on. If None, an axis is created
            from the current figure. Defaults to None.
        plot_kwargs: kwargs for plotting with matplotlib.pyplot.scatter (plt.scatter)
    """

    _, ax = _check_and_create_fig_ax(ax=ax)

    plot_scatter_points(
        scatter_points=scatter_points,
        direction_x=pca_object.components[pca_object.components_ids[0]],
        direction_y=pca_object.components[pca_object.components_ids[1]],
        shift=pca_object.mean,
        ax=ax,
        **plot_kwargs,
    )


def plot_line_through_points_on_pca(
    points: ArrayOfParameterVectors,
    pca_object: PCAobject,
    ax: Optional[matplotlib.axes.Axes] = None,
    **plot_kwargs,
):
    """Wrapper function around plot_line_through_points to simplify plotting a on PCA scan.

    Args:
        points: Parameter points to be projected on a PCA scan and plotted with a line.
        pca_object: PCAobject the contains a fitted PCA object, corresponding points
            and the components of interest.
        ax: Matplotlib axis to perform the plot on. If None, an axis is created
            from the current figure. Defaults to None.
        plot_kwargs: kwargs for plotting with matplotlib.pyplot.plot (plt.plot)
    """

    _, ax = _check_and_create_fig_ax(ax=ax)

    plot_line_through_points(
        points=points,
        direction_x=pca_object.components[pca_object.components_ids[0]],
        direction_y=pca_object.components[pca_object.components_ids[1]],
        shift=pca_object.mean,
        ax=ax,
        **plot_kwargs,
    )


def plot_pca_landscape(
    scan_result: Scan2DResult,
    pca_object: PCAobject,
    fig: Optional[matplotlib.figure.Figure] = None,
    ax: Optional[matplotlib.axes.Axes] = None,
    **plot_kwargs,
):
    """Wrapper function around plot_2D_scan_result to simplify plotting a on PCA scan.

    Args:
        scan_result: Scan2DResult object from a performed 2D scan.
        pca: sklearn PCA object fitted on the optimization trajectory.
        fig: Matplotlib figure to perform the plot on. If None, a new figure
            and axis are created. Defaults to None.
        ax: Matplotlib axis to perform the plot on. If None, a new axis is created
            from the current figure. Defaults to None.
        plot_kwargs: kwargs for plotting with matplotlib.pyplot.pcolormesh
            (plt.pcolormesh)
    """
    fig, ax = _check_and_create_fig_ax(fig=fig, ax=ax)

    plot_2D_scan_result(
        scan_result,
        fig=fig,
        ax=ax,
        **plot_kwargs,
    )

    component1 = pca_object.components_ids[0]
    component2 = pca_object.components_ids[1]
    ax.set_xlabel(
        "{}. component explains {:.2f}% of variance".format(
            component1, pca_object.pca.explained_variance_ratio_[component1] * 100
        )
    )
    ax.set_ylabel(
        "{}. component explains {:.2f}% of variance".format(
            component2, pca_object.pca.explained_variance_ratio_[component2] * 100
        )
    )



import numpy as np
import pytest

from orqviz.geometric import (
    direction_linspace,
    get_coordinates_on_direction,
    get_random_normal_vector,
    get_random_orthonormal_vector,
    relative_periodic_trajectory_wrap,
    relative_periodic_wrap,
    uniformly_distribute_trajectory,
)


@pytest.mark.parametrize("dimension", [2, 5, 100, 1000, (2, 3), (11, 4)])
def test_get_random_normal_vector(dimension):
    vector = get_random_normal_vector(dimension=dimension)
    if isinstance(dimension, int):
        assert len(vector) == dimension
    elif isinstance(dimension, tuple):
        assert vector.shape == dimension
    assert np.isclose(np.linalg.norm(vector), 1.0)


@pytest.mark.parametrize("dimension", [2, 5, 100, 1000, (2, 3), (11, 4)])
def test_get_random_orthonormal_vector(dimension):
    vector_1 = get_random_normal_vector(dimension=dimension)
    vector_2 = get_random_orthonormal_vector(vector_1)
    if isinstance(dimension, int):
        assert len(vector_1) == len(vector_2) == dimension
    elif isinstance(dimension, tuple):
        assert vector_1.shape == vector_2.shape == dimension
    assert np.isclose(np.linalg.norm(vector_2), 1.0)
    assert np.isclose(np.dot(vector_1.flatten(), vector_2.flatten()), 0.0)


@pytest.mark.parametrize(
    "reference_point,point_to_be_wrapped,period,target_point",
    [
        [np.array([0, 0]), np.array([1, 0]), np.pi, np.array([1, 0])],
        [np.array([0, 0]), np.array([1, 1]), np.pi, np.array([1, 1])],
        [np.array([0, 0]), np.array([1, 0]), 0.8, np.array([0.2, 0])],
        [np.array([0, 0]), np.array([1, 0]), 0.1, np.array([0, 0])],
        [np.array([0, 0]), np.array([1, 0]), 0.11, np.array([0.01, 0])],
        [
            np.array([0, 0]),
            np.array([1, 1]),
            np.sqrt(2) / 2,
            np.array([1 - np.sqrt(2) / 2, 1 - np.sqrt(2) / 2]),
        ],
        [np.array([0, 0]), np.array([3, 4]), 2.5, np.array([0.5, -1])],
        [np.array([1, 1]), np.array([4, 5]), 5, np.array([-1, 0])],
        [
            np.array([[1, 1], [2, 2]]),
            np.array([[4, 5], [6, 7]]),
            3,
            np.array([[1, 2], [3, 1]]),
        ],
        [
            np.array([[1, 1], [2, 2]]),
            np.array([[1.2, -3], [0, 7.1]]),
            1.4,
            np.array([[1.2, 1.2], [1.4, 1.5]]),
        ],
    ],
)
def test_relative_periodic_wrap(
    reference_point, point_to_be_wrapped, period, target_point
):
    resulting_point = relative_periodic_wrap(
        reference_point, point_to_be_wrapped, period
    )
    assert np.allclose(resulting_point, target_point)


@pytest.mark.parametrize(
    "reference_point,trajectory,period,target_trajectory",
    [
        [
            np.array([0, 0]),
            np.array([[4, 0], [3, 0], [2, 0], [1, 0]]),
            np.pi,
            np.array([[4, 0], [3, 0], [2, 0], [1, 0]]),
        ],
        [
            np.array([0, 0]),
            np.array([[6, 0], [5, 0], [4, 0], [3, 0]]),
            2.5,
            np.array([[3.5, 0], [2.5, 0], [1.5, 0], [0.5, 0]]),
        ],
    ],
)
def test_relative_periodic_trajectory_wrap(
    reference_point, trajectory, period, target_trajectory
):
    resulting_trajectory = relative_periodic_trajectory_wrap(
        reference_point, trajectory, period
    )
    assert np.allclose(resulting_trajectory, target_trajectory)


@pytest.mark.parametrize(
    "reference_point,trajectory,period",
    [
        [
            np.array([0, 0]),
            np.array([[4, 0], [3, 0], [2, 0], [1, 0]]),
            0.1,
        ],
    ],
)
def test_relative_periodic_trajectory_wrap_fails_for_wrong_input(
    reference_point, trajectory, period
):
    with pytest.raises(ValueError):
        _ = relative_periodic_trajectory_wrap(reference_point, trajectory, period)


@pytest.mark.parametrize(
    "points,direction,origin,in_units_of_direction,target_points",
    [
        [
            np.array([[0, 0], [1, 1]]),
            np.array([2, 0]),
            np.array([0, 0]),
            False,
            np.array([0, 1]),
        ],
        [
            np.array([[0, 0], [1, 1]]),
            np.array([2, 0]),
            np.array([0, 0]),
            True,
            np.array([0, 0.5]),
        ],
        [
            np.array([[0, 0], [1, 1]]),
            np.array([2, 0]),
            None,
            False,
            np.array([0, 1]),
        ],
        [
            np.array([[0, 0], [1, 1]]),
            np.array([1, 0]),
            np.array([1, 1]),
            False,
            np.array([-1, 0]),
        ],
    ],
)
def test_get_coordinates_on_direction(
    points, direction, origin, in_units_of_direction, target_points
):
    transformed_points = get_coordinates_on_direction(
        points, direction, origin, in_units_of_direction
    )
    assert np.allclose(transformed_points, target_points)


@pytest.mark.parametrize(
    "origin,direction,n_points,endpoints,target_vectors",
    [
        [
            np.array([0, 0]),
            np.array([0, 1]),
            5,
            (-2, 2),
            np.array([[0, -2], [0, -1], [0, 0], [0, 1], [0, 2]]),
        ],
        [
            np.array([1, 1]),
            np.array([-2, 1]),
            4,
            (-1, 1),
            np.array([[3, 0], [5 / 3, 2 / 3], [1 / 3, 4 / 3], [-1, 2]]),
        ],
    ],
)
def test_direction_linspace(origin, direction, n_points, endpoints, target_vectors):
    result = direction_linspace(origin, direction, n_points, endpoints)
    assert np.allclose(result, target_vectors)


@pytest.mark.parametrize(
    "trajectory,n_points,target_trajectory",
    [
        [
            np.array([[1, 0], [2, 0], [3, 0], [4, 0]]),
            4,
            np.array([[1, 0], [2, 0], [3, 0], [4, 0]]),
        ],
        [
            np.array([[1, 0], [2, 0], [3, 0], [7, 0]]),
            4,
            np.array([[1, 0], [3, 0], [5, 0], [7, 0]]),
        ],
        [
            np.array([[0, 0], [3, 0], [3, 3]]),
            5,
            np.array([[0, 0], [1.5, 0], [3, 0], [3, 1.5], [3, 3]]),
        ],
    ],
)
def test_uniformly_distribute_trajectory(trajectory, n_points, target_trajectory):
    transformed_trajectory = uniformly_distribute_trajectory(trajectory, n_points)
    assert np.allclose(transformed_trajectory, target_trajectory)



import os

import numpy as np
import pytest

from orqviz.elastic_band.data_structures import Chain
from orqviz.hessians import get_Hessian
from orqviz.scans import perform_1D_scan, perform_2D_scan
from orqviz.utils import OrqVizObject, load_viz_object, save_viz_object


def SUM_OF_SINES(params):
    return np.sum(np.sin(params))


def test_saving_and_loading_datatypes():
    origin = np.random.rand(2)
    direction_x = np.random.rand(2)
    direction_y = np.random.rand(2)
    n_steps_x = 2

    scan1d = perform_1D_scan(
        loss_function=SUM_OF_SINES,
        origin=origin,
        direction=direction_x,
        n_steps=n_steps_x,
    )

    scan2d = perform_2D_scan(
        origin=origin,
        loss_function=SUM_OF_SINES,
        direction_x=direction_x,
        direction_y=direction_y,
        n_steps_x=n_steps_x,
    )

    hessian = get_Hessian(params=origin, loss_function=SUM_OF_SINES)

    chain = Chain(np.linspace(origin, origin + direction_x, num=5))

    for data_object in [scan1d, scan2d, hessian, chain]:
        with pytest.warns(DeprecationWarning):
            save_viz_object(data_object, "test")
        with pytest.warns(DeprecationWarning):
            loaded_data_object = load_viz_object("test")
        os.remove("test")
        assert isinstance(loaded_data_object, OrqVizObject.__args__)
        assert type(loaded_data_object) == type(data_object)



from typing import Callable, Optional

import numpy as np

from .aliases import DirectionVector, GradientFunction, LossFunction, ParameterVector


def calculate_full_gradient(
    params: ParameterVector,
    loss_function: LossFunction,
    gradient_function: Optional[GradientFunction] = None,
    stochastic: bool = False,
    eps: float = 1e-3,
) -> ParameterVector:
    """Function to calculate a full gradient vector of partial derivatives
        of a loss function with respect to each entry of a parameter vector.

    Args:
        params: Parameter vector at which to calculate the gradient.
        loss_function: Function with respect to which the gradient is calculated.
            It is not required when a gradient_function is passed. It must receive
            only a numpy.ndarray of parameters, and return a real number.
            If your function requires more arguments, consider using the
            'LossFunctionWrapper' class from 'orqviz.loss_function'.
        gradient_function: Gradient function to calculate the partial derivative
            of the loss function with respect to a direction.
            If None, a numerical gradient is computed with the passed loss function.
            Defaults to None.
        stochastic: Flag to indicate whether a stochastic gradient is computed in
            a randomly sampled direction.
            If True, the specified gradient function must support this.
            Defaults to False.
        eps: Stencil for numerical gradient calculation. If the loss function is noisy,
            this value needs to be increase, e.g to 0.1. Defaults to 1e-3.
    """

    if loss_function is None and gradient_function is None:
        raise Exception(
            """
            'loss_function' and 'gradient_function' cannot both be 'None'
            """
        )

    if gradient_function is None:

        def _gradient_function(pars: ParameterVector, direction: np.ndarray) -> float:
            return numerical_gradient(
                pars, direction, loss_function=loss_function, eps=eps
            )

        gradient_function = _gradient_function

    grad = np.zeros_like(params)

    if stochastic:
        direction = np.random.choice([-1.0, 1.0], size=np.shape(params))
        grad_value = gradient_function(params, direction)
        grad = grad + grad_value * direction
    else:
        flat_params = params.flatten()
        for i in range(len(flat_params)):
            direction = np.zeros_like(flat_params)
            direction[i] = 1.0
            direction = direction.reshape(params.shape)
            grad_value = gradient_function(params, direction)
            grad = grad + grad_value * direction
    return grad


def numerical_gradient(
    x: ParameterVector,
    direction: DirectionVector,
    loss_function: LossFunction,
    eps: float,
) -> float:
    """Function to calculate a numerical gradient of a loss function at point x
    with respect to a specified direction."""
    x_plus = x + eps * direction
    f_plus = loss_function(x_plus)

    x_minus = x - eps * direction
    f_minus = loss_function(x_minus)
    return (f_plus - f_minus) / (2 * eps)



import matplotlib.pyplot as plt
import numpy as np
import pytest

from orqviz.pca import (
    get_pca,
    perform_2D_pca_scan,
    plot_line_through_points_on_pca,
    plot_optimization_trajectory_on_pca,
    plot_pca_landscape,
    plot_scatter_points_on_pca,
)
from orqviz.scans.data_structures import Scan2DResult


def COST_FUNCTION(params):
    return np.sum(np.sin(params)) + np.sum(params**2) + 3 * params[1] - 10 * params[2]


@pytest.fixture
def all_points():
    return np.random.rand(98, 9)


@pytest.fixture
def pca(all_points):

    return get_pca(all_points, components_ids=(0, 1))


@pytest.fixture
def scan_results(pca):
    n_steps_x = 100
    n_steps_y = 20

    return perform_2D_pca_scan(
        pca_object=pca,
        loss_function=COST_FUNCTION,
        n_steps_x=n_steps_x,
        n_steps_y=n_steps_y,
    )


def test_plot_pca_landscape(scan_results, pca):
    plot_pca_landscape(scan_results, pca_object=pca)
    fig, ax = plt.subplots(1, 1)
    plot_pca_landscape(scan_results, pca_object=pca, fig=fig, ax=ax)


def test_plot_optimization_trajectory_on_pca(all_points, pca):
    plot_optimization_trajectory_on_pca(all_points, pca_object=pca)
    fig, ax = plt.subplots(1, 1)
    plot_optimization_trajectory_on_pca(all_points, pca_object=pca, ax=ax)


def test_plot_scatter_points_on_pca(all_points, pca):
    plot_scatter_points_on_pca(all_points, pca_object=pca)
    fig, ax = plt.subplots(1, 1)
    plot_scatter_points_on_pca(all_points, pca_object=pca, ax=ax)


def test_plot_line_through_points_on_pca(all_points, pca):
    plot_line_through_points_on_pca(all_points, pca_object=pca)
    fig, ax = plt.subplots(1, 1)
    plot_line_through_points_on_pca(all_points, pca_object=pca, ax=ax)



import os

import matplotlib.pyplot as plt
import numpy as np

from orqviz.elastic_band import Chain, plot_all_chains_losses, run_AutoNEB, run_NEB
from orqviz.io import load_viz_object, save_viz_object


def SUM_OF_SINES(params):
    return np.sum(np.sin(params))


def test_NEB():
    pivots = np.array([[t, 1 - t] for t in np.linspace(0, 1, 10)])
    init_chain = Chain(pivots=pivots)

    n_iters = 10
    all_chains = run_NEB(
        init_chain,
        loss_function=SUM_OF_SINES,
        full_gradient_function=None,
        n_iters=n_iters,
    )

    assert len(all_chains) == n_iters + 1
    for chain in all_chains:
        np.testing.assert_array_almost_equal(chain.pivots[0], np.array([0, 1]))
        np.testing.assert_array_almost_equal(chain.pivots[-1], np.array([1, 0]))


def test_AutoNEB():
    pivots = np.array([[t, 1 - t] for t in np.linspace(0, 1, 10)])
    init_chain = Chain(pivots=pivots)

    n_cycles = 4
    max_new_pivots = 2
    all_chains = run_AutoNEB(
        init_chain,
        loss_function=SUM_OF_SINES,
        full_gradient_function=None,
        n_cycles=n_cycles,
        n_iters_per_cycle=10,
        max_new_pivots=max_new_pivots,
    )

    for chain in all_chains:
        np.testing.assert_array_almost_equal(chain.pivots[0], np.array([0, 1]))
        np.testing.assert_array_almost_equal(chain.pivots[-1], np.array([1, 0]))

    save_viz_object(all_chains[-1], "test")
    loaded_chain = load_viz_object("test")
    os.remove("test")
    np.testing.assert_array_almost_equal(loaded_chain.pivots[0], np.array([0, 1]))
    np.testing.assert_array_almost_equal(loaded_chain.pivots[-1], np.array([1, 0]))
    np.testing.assert_array_almost_equal(loaded_chain.pivots, all_chains[-1].pivots)
    assert (
        all_chains[0].n_pivots
        <= all_chains[-1].n_pivots
        <= all_chains[0].n_pivots + (n_cycles - 1) * max_new_pivots
    )



import os

import matplotlib.pyplot as plt
import numpy as np
import pytest

from orqviz.geometric import get_random_normal_vector, get_random_orthonormal_vector
from orqviz.io import load_viz_object, save_viz_object
from orqviz.scans import (
    perform_1D_interpolation,
    perform_1D_scan,
    perform_2D_interpolation,
)
from orqviz.scans.data_structures import Scan1DResult, Scan2DResult
from orqviz.scans.scans_2D import perform_2D_scan


def SUM_OF_SINES(params):
    return np.sum(np.sin(params))


def test_1D_scan():
    origin = np.array([0.0, 0.0])
    direction_x = np.array([1.0, 0.0])
    end_points_x = (-np.pi, np.pi)
    n_steps_x = 100

    scan_1d = perform_1D_scan(
        origin=origin,
        loss_function=SUM_OF_SINES,
        direction=direction_x,
        n_steps=n_steps_x,
        end_points=end_points_x,
    )
    save_viz_object(scan_1d, "test")
    loaded_scan1d = load_viz_object("test")
    os.remove("test")

    for scan_results in [scan_1d, loaded_scan1d]:
        assert isinstance(scan_results, Scan1DResult)
        assert len(scan_results.values) == n_steps_x
        np.testing.assert_equal(scan_results.direction, direction_x)
        assert scan_results.params_list.shape == (n_steps_x, 2)

        np.testing.assert_equal(
            scan_results.params_list[0],
            origin + direction_x * end_points_x[0],
        )
        np.testing.assert_equal(
            scan_results.params_list[-1],
            origin + direction_x * end_points_x[1],
        )


def test_1D_scan_works_with_default_direction():
    origin = np.array([0.0, 0.0])
    n_steps_x = 31

    scan_1d = perform_1D_scan(
        origin=origin,
        loss_function=SUM_OF_SINES,
    )

    assert isinstance(scan_1d, Scan1DResult)
    assert len(scan_1d.values) == n_steps_x
    assert len(scan_1d.direction) == len(origin)
    assert scan_1d.params_list.shape == (n_steps_x, 2)


def test_1D_interpolation():
    point1 = np.array([1, 0])
    point2 = np.array([0, 1])
    direction_x = point2 - point1
    end_points_x = (0, 1)
    n_steps_x = 23

    scan_1d = perform_1D_interpolation(
        point_1=point1,
        point_2=point2,
        loss_function=SUM_OF_SINES,
        n_steps=n_steps_x,
        end_points=end_points_x,
    )
    save_viz_object(scan_1d, "test")
    loaded_scan1d = load_viz_object("test")
    os.remove("test")

    for scan_results in [scan_1d, loaded_scan1d]:
        assert isinstance(scan_results, Scan1DResult)
        assert len(scan_results.values) == n_steps_x
        np.testing.assert_equal(scan_results.direction, direction_x)
        assert scan_results.params_list.shape == (n_steps_x, 2)

        np.testing.assert_equal(
            scan_results.params_list[0],
            point1,
        )
        np.testing.assert_equal(
            scan_results.params_list[-1],
            point2,
        )


def test_perform_2D_scan_in_2D_space():
    origin = np.array([0, 0])
    direction_x = np.array([1, 0])
    direction_y = np.array([0, 1])
    end_points_x = (-np.pi, np.pi)
    end_points_y = (-np.pi / 2, np.pi / 2)
    n_steps_x = 100
    n_steps_y = 10

    scan_2d = perform_2D_scan(
        origin=origin,
        loss_function=SUM_OF_SINES,
        direction_x=direction_x,
        direction_y=direction_y,
        n_steps_x=n_steps_x,
        n_steps_y=n_steps_y,
        end_points_x=end_points_x,
        end_points_y=end_points_y,
    )
    save_viz_object(scan_2d, "test")
    loaded_scan2d = load_viz_object("test")
    os.remove("test")

    for scan_results in [scan_2d, loaded_scan2d]:
        assert isinstance(scan_results, Scan2DResult)
        assert scan_results.values.shape == (n_steps_y, n_steps_x)
        np.testing.assert_equal(scan_results.direction_x, direction_x)
        np.testing.assert_equal(scan_results.direction_y, direction_y)
        assert scan_results.params_grid.shape == (n_steps_y, n_steps_x, 2)

        for i, j in [(0, 0), (0, 1), (1, 0), (0, 1)]:
            np.testing.assert_equal(
                scan_results.params_grid[-i][-j],
                origin + end_points_x[j] * direction_x + end_points_y[i] * direction_y,
            )


def test_2D_scan_works_with_default_direction():
    origin = np.array([0.0, 0.0])
    n_steps_x = 20
    n_steps_y = 20

    scan_2d = perform_2D_scan(
        origin=origin,
        loss_function=SUM_OF_SINES,
    )

    assert isinstance(scan_2d, Scan2DResult)
    assert len(scan_2d.values) == n_steps_x
    assert len(scan_2d.direction_x) == len(scan_2d.direction_y) == len(origin)
    assert scan_2d.params_grid.shape == (n_steps_y, n_steps_x, 2)


def test_perform_2D_scan_in_5D_space():
    origin = np.zeros(5)
    direction_x = np.array([0, 0, 0, 1, 0])
    direction_y = np.array([0, 0, 0, 0, 1])
    end_points_x = (-np.pi, np.pi)
    end_points_y = (-np.pi / 2, np.pi / 2)
    n_steps_x = 100
    n_steps_y = 20

    scan_2d = perform_2D_scan(
        origin=origin,
        loss_function=SUM_OF_SINES,
        direction_x=direction_x,
        direction_y=direction_y,
        n_steps_x=n_steps_x,
        n_steps_y=n_steps_y,
        end_points_x=end_points_x,
        end_points_y=end_points_y,
    )

    save_viz_object(scan_2d, "test")
    loaded_scan2d = load_viz_object("test")
    os.remove("test")

    for scan_results in [scan_2d, loaded_scan2d]:
        assert isinstance(scan_results, Scan2DResult)
        assert scan_results.values.shape == (n_steps_y, n_steps_x)
        np.testing.assert_equal(scan_results.direction_x, direction_x)
        np.testing.assert_equal(scan_results.direction_y, direction_y)
        assert scan_results.params_grid.shape == (n_steps_y, n_steps_x, 5)

        for i, j in [(0, 0), (0, 1), (1, 0), (0, 1)]:
            np.testing.assert_equal(
                scan_results.params_grid[-i][-j],
                origin + end_points_x[j] * direction_x + end_points_y[i] * direction_y,
            )


def test_perform_2D_scan_with_ND_array():
    param_shape = (9, 4)
    origin = np.zeros(param_shape)
    direction_x = get_random_normal_vector(param_shape)
    direction_y = get_random_orthonormal_vector(direction_x)
    end_points_x = (-np.pi, np.pi)
    end_points_y = (-np.pi / 2, np.pi / 2)
    n_steps_x = 100
    n_steps_y = 10

    scan_2d = perform_2D_scan(
        origin=origin,
        loss_function=SUM_OF_SINES,
        direction_x=direction_x,
        direction_y=direction_y,
        n_steps_x=n_steps_x,
        n_steps_y=n_steps_y,
        end_points_x=end_points_x,
        end_points_y=end_points_y,
    )
    save_viz_object(scan_2d, "test")
    loaded_scan2d = load_viz_object("test")
    os.remove("test")

    for scan_results in [scan_2d, loaded_scan2d]:
        assert isinstance(scan_results, Scan2DResult)
        assert scan_results.values.shape == (n_steps_y, n_steps_x)
        np.testing.assert_equal(scan_results.direction_x, direction_x)
        np.testing.assert_equal(scan_results.direction_y, direction_y)
        assert scan_results.params_grid.shape == (n_steps_y, n_steps_x, *param_shape)

        for i, j in [(0, 0), (0, 1), (1, 0), (0, 1)]:
            np.testing.assert_equal(
                scan_results.params_grid[-i][-j],
                origin + end_points_x[j] * direction_x + end_points_y[i] * direction_y,
            )


def test_interpolation_2D_in_2D_space():
    point_1 = np.array([1, 0])
    point_2 = np.array([0, 2])
    end_points_x = (-0.5, 1.5)
    end_points_y = (-0.5, 0.5)
    n_steps_x = 100
    n_steps_y = 10

    scan_2d = perform_2D_interpolation(
        point_1=point_1,
        point_2=point_2,
        loss_function=SUM_OF_SINES,
        n_steps_x=n_steps_x,
        n_steps_y=n_steps_y,
        end_points_x=end_points_x,
        end_points_y=end_points_y,
    )

    save_viz_object(scan_2d, "test")
    loaded_scan2d = load_viz_object("test")
    os.remove("test")

    for scan_results in [scan_2d, loaded_scan2d]:
        assert isinstance(scan_results, Scan2DResult)
        assert scan_results.values.shape == (n_steps_y, n_steps_x)
        np.testing.assert_equal(scan_results.direction_x, point_2 - point_1)
        assert np.dot(
            scan_results.direction_x, scan_results.direction_y
        ) == pytest.approx(0)
        assert scan_results.params_grid.shape == (n_steps_y, n_steps_x, 2)



from typing import Callable, List, Optional

import matplotlib
import numpy as np

from ..aliases import LossFunction, ParameterVector
from ..plot_utils import _check_and_create_fig_ax
from ..scans import eval_points_on_path
from .neb import Chain


def plot_all_chains_losses(
    all_chains: List[Chain],
    loss_function: LossFunction,
    ax: Optional[matplotlib.axes.Axes] = None,
    **plot_kwargs,
) -> None:
    """Function to plot

    Args:
        all_chains: List of Chains to evaluate the loss on.
        loss_function: Function to evaluate the chain pivots on. It must receive only a
            numpy.ndarray of parameters, and return a real number.
            If your function requires more arguments, consider using the
            'LossFunctionWrapper' class from 'orqviz.loss_function'.
        ax: Matplotlib axis to plot on. If None, a new axis is created
            from the current figure. Defaults to None.
        plot_kwargs: kwargs for plotting with matplotlib.pyplot.plot (plt.plot)
    """
    _, ax = _check_and_create_fig_ax(ax=ax)

    for chain in all_chains:
        losses = eval_points_on_path(chain.pivots, loss_function)
        inter_points = chain.get_weights()
        ax.plot(inter_points, losses, **plot_kwargs)

    ax.set_xlabel("Chain Position")
    ax.set_ylabel("Loss Value")



from typing import Callable, List, Optional, Tuple

import matplotlib
import numpy as np

from ..plot_utils import _check_and_create_fig_ax
from ..scans import Scan1DResult, perform_1D_scan
from ..scans.plots import plot_1D_scan_result


def plot_1D_hessian_eigenvector_scan_result(
    list_of_scans: List[Scan1DResult],
    eigenvalues: Optional[np.ndarray] = None,
    ax: Optional[matplotlib.axes.Axes] = None,
    **plot_kwargs,
):
    """Function to plot 1D scans along all eigenvecor directions of a Hessian

    Args:
        list_of_scans: List of Scan1DResult that
        ax: Matplotlib axis to perform the plot on.
            If None, a new axis is created from the current figure. Defaults to None.
        n_points: Number of points to evaluate the loss along each direction.
            Defaults to 31.
        endpoints: End points for scan along each direction.
            Defaults to (-np.pi, np.pi).
        plot_kwargs: kwargs for plotting with matplotlib.pyplot.plot (plt.plot)
    """
    _, ax = _check_and_create_fig_ax(ax=ax)

    for ii in range(len(list_of_scans)):
        plot_1D_scan_result(
            list_of_scans[ii],
            ax,
            label="{:.3f}".format(eigenvalues[ii]) if eigenvalues is not None else "",
            **plot_kwargs,
        )

    ax.set_xlabel("Eigenvector direction")
    ax.set_ylabel("Loss Value")
    if eigenvalues is not None:
        ax.legend()



from typing import Optional, Tuple

import numpy as np

from ..aliases import (
    DirectionVector,
    GridOfParameterVectors,
    LossFunction,
    ParameterVector,
)
from ..geometric import (
    get_random_normal_vector,
    get_random_orthonormal_vector,
    relative_periodic_wrap,
)
from .data_structures import Scan2DResult
from .evals import eval_points_on_grid


def perform_2D_interpolation(
    point_1: ParameterVector,
    point_2: ParameterVector,
    loss_function: LossFunction,
    direction_y: Optional[DirectionVector] = None,
    n_steps_x: int = 20,
    n_steps_y: Optional[int] = None,
    end_points_x: Tuple[float, float] = (-0.5, 1.5),
    end_points_y: Tuple[float, float] = (-0.5, 0.5),
    parameter_period: Optional[float] = None,
    verbose: bool = False,
) -> Scan2DResult:
    """Function to perform a 2D scan to interpolate between two points.

    Args:
        point_1: First point of the interpolation.
        point_2: Second point of the interpolation.
        loss_function: Function to perform the scan on. It must receive only a
            numpy.ndarray of parameters, and return a real number.
            If your function requires more arguments, consider using the
            'LossFunctionWrapper' class from 'orqviz.loss_function'.
        direction_y: Second scan direction for the 2D scan where first direction
            is the interpolation vector of the points. If None, it's chosen at random.
        n_steps_x: Number of points evaluated along the x-direction. Defaults to 20.
        n_steps_y: Number of points evaluated along the y-direction.
            If None, set value to n_steps_y. Defaults to None.
        end_points_x: Range of scan along the x-direction in units
            of the interpolation vector. Defaults to (-0.5, 1.5).
        end_points_y: Range of scan along the y-direction in units
            of the interpolation vector. Defaults to (-0.5, 0.5).
        parameter_period: Optional period of the parameters to scan the shortest
            interpolated path between the points.
            If None, interpolation per parameter happens along the real number line.
            Defaults to None.
        verbose: Flag for printing progress. Defaults to False.
    """
    if n_steps_y is None:
        n_steps_y = n_steps_x

    if parameter_period:
        point_2 = relative_periodic_wrap(point_1, point_2, parameter_period)

    direction_x = point_2 - point_1

    if direction_y is None:
        direction_y = get_random_orthonormal_vector(direction_x) * np.linalg.norm(
            direction_x
        )

    return perform_2D_scan(
        origin=point_1,
        loss_function=loss_function,
        direction_x=direction_x,
        direction_y=direction_y,
        n_steps_x=n_steps_x,
        n_steps_y=n_steps_y,
        end_points_x=end_points_x,
        end_points_y=end_points_y,
        verbose=verbose,
    )


def perform_2D_scan(
    origin: ParameterVector,
    loss_function: LossFunction,
    direction_x: Optional[DirectionVector] = None,
    direction_y: Optional[DirectionVector] = None,
    n_steps_x: int = 20,
    n_steps_y: Optional[int] = None,
    end_points_x: Tuple[float, float] = (-1, 1),
    end_points_y: Optional[Tuple[float, float]] = None,
    verbose: bool = False,
) -> Scan2DResult:
    """Function to perform a 2D scan around a point on a loss function.

    Args:
        origin: Origin point of the 2D scan.
        loss_function: Function to perform the scan on. It must receive only a
            numpy.ndarray of parameters, and return a real number.
            If your function requires more arguments, consider using the
            'LossFunctionWrapper' class from 'orqviz.loss_function'.
        direction_x: x-direction vector for scan. Has same shape as origin.
            If None, a random unit vector is sampled. Defaults to None.
        direction_y: y-direction vector for scan. Has same shape as origin.
            If None, a random unit vector is sampled. Defaults to None.
        n_steps_x: Number of points evaluated along the x-direction. Defaults to 20.
        n_steps_y: Number of points evaluated along the y-direction.
            If None, set value to n_steps_x. Defaults to None.
        end_points_x: Range of scan along the x-direction in units of direction_x.
            Defaults to (-1, 1).
        end_points_y: Range of scan along the y-direction in units of direction_y.
            Defaults to (-1, 1).
        verbose: Flag for printing progress. Defaults to False.
    """
    if direction_x is None:
        direction_x = get_random_normal_vector(np.shape(origin))
    if direction_y is None:
        direction_y = get_random_orthonormal_vector(direction_x) * np.linalg.norm(
            direction_x
        )
    if n_steps_y is None:
        n_steps_y = n_steps_x

    if end_points_y is None:
        end_points_y = end_points_x

    interpolation_steps_x = np.linspace(
        end_points_x[0], end_points_x[1], num=n_steps_x, endpoint=True
    )
    interpolation_steps_y = np.linspace(
        end_points_y[0], end_points_y[1], num=n_steps_y, endpoint=True
    )
    x_values, y_values = np.meshgrid(interpolation_steps_x, interpolation_steps_y)

    # We acknowledge the fact that this method is not the most intuitive one,
    # but it does the job, so we decided to leave it here.
    params_grid: GridOfParameterVectors = (
        origin
        + np.tensordot(x_values, direction_x, axes=-1)
        + np.tensordot(y_values, direction_y, axes=-1)
    )

    loss_grid = eval_points_on_grid(
        params_grid, loss_function=loss_function, n_reps=1, verbose=verbose
    )
    return Scan2DResult(
        params_grid=params_grid,
        direction_x=direction_x,
        direction_y=direction_y,
        values=loss_grid,
        origin=origin,
    )



import numpy as np

from orqviz.loss_function import LossFunctionWrapper


def mock_function(params, b, c):
    return np.sum(params * b + c)


class TestLossFunctionWrapper:
    def test_if_wraps_args(self):
        b = 10
        c = 2
        params = np.array([1, 2, 3, 4])
        target_value = mock_function(params, b, c)

        loss_function = LossFunctionWrapper(mock_function, b, c)

        result = loss_function(params=params)
        assert result == target_value

    def test_if_wraps_kwargs(self):
        b = 10
        c = 2
        params = np.array([1, 2, 3, 4])
        target_value = mock_function(params, b, c)

        loss_function = LossFunctionWrapper(mock_function, b=b, c=c)

        result = loss_function(params=params)
        assert result == target_value

    def test_if_wraps_both_args_and_kwargs(self):
        b = 10
        c = 2
        params = np.array([1, 2, 3, 4])
        target_value = mock_function(params, b, c)

        loss_function = LossFunctionWrapper(mock_function, b, c=c)

        result = loss_function(params=params)

        assert result == target_value

    def test_tracks_average_call_time(self):
        b = 10
        c = 2
        call_count = 10

        loss_function = LossFunctionWrapper(mock_function, b, c=c)
        for i in range(call_count):
            _ = loss_function(params=np.random.random(i))

        assert loss_function.average_call_time is not None

    def test_tracks_call_count(self):
        b = 10
        c = 2
        call_count = 10

        loss_function = LossFunctionWrapper(mock_function, b, c=c)
        for i in range(call_count):
            _ = loss_function(params=np.random.random(i))

        assert loss_function.call_count == call_count

    def test_tracks_min_value(self):
        b = 10
        c = 2
        call_count = 10

        loss_function = LossFunctionWrapper(mock_function, b, c=c)
        min_value = np.inf
        for i in range(call_count):
            value = loss_function(params=np.random.random(i))
            if value < min_value:
                min_value = value

        assert loss_function.min_value == min_value

    def test_reset(self):
        b = 10
        c = 2
        call_count = 10

        loss_function = LossFunctionWrapper(mock_function, b, c=c)
        for i in range(call_count):
            _ = loss_function(params=np.random.random(i))

        loss_function.reset()

        assert loss_function.average_call_time is None
        assert loss_function.call_count == 0
        assert loss_function.min_value is None
